###########################################################################################
# Asteroid support code
#
# (c) Lutz Hamel, University of Rhode Island
###########################################################################################

import re

from asteroid.state import state

###########################################################################################
def dump_AST(node):
    '''
    this function will print any AST that follows the

         (TYPE [, child1, child2,...])

    tuple format for tree nodes.
    '''
    _dump_AST(node)
    print('')

def _dump_AST(node, level=0):

    if isinstance(node, tuple):
        _indent(level)
        nchildren = len(node) - 1

        print("(%s" % node[0], end='')

        if nchildren > 0:
            print(" ", end='')

        for c in range(nchildren):
            _dump_AST(node[c+1], level+1)
            if c != nchildren-1:
                print(' ', end='')

        print(")", end='')

    elif isinstance(node, list):
        _indent(level)
        print("[", end='')

        nchildren = len(node)

        if nchildren > 0:
            print(" ", end='')

        for c in range(nchildren):
            _dump_AST(node[c], level+1)
            if c != nchildren-1:
                print(' ', end='')

        print("]", end='')

    else:
        print("%s" % str(node), end='')

def _indent(level):
    print('')
    for i in range(level):
        print('  |',end='')


###########################################################################################
def assert_match(input, expected):
    nomatch = False

    if isinstance(expected, list):
        if input not in expected:
            nomatch = True
    elif input != expected:
        nomatch = True

    if nomatch:
        raise ValueError(
            "Internal Error: pattern assert failed: expected '{}' but got '{}'".
            format(expected, input))

###########################################################################################
def promote(type1, type2):
    '''
    type promotion table for builtin primitive types.  this table implements the
    type hierarchies

                 boolean < integer < real < string
                 list < string
                 tuple < string
                 none
    '''

    if type1 == 'string' and type2 in['string','real','integer','list','tuple','boolean']:
        return 'string'
    if type2 == 'string' and type1 in['string','real','integer','list','tuple','boolean']:
        return 'string'
    elif type1 == 'real' and type2 in ['real', 'integer', 'boolean']:
        return 'real'
    elif type2 == 'real' and type1 in ['real', 'integer', 'boolean']:
        return 'real'
    elif type1 == 'integer' and type2 in ['integer', 'boolean']:
        return 'integer'
    elif type2 == 'integer' and type1 in ['integer', 'boolean']:
        return 'integer'
    elif type1 == 'boolean' and type2 == 'boolean':
        return 'boolean'
    elif type1 == 'list' and type2 == 'list':
        return 'list'
    elif type1 == 'tuple' and type2 == 'tuple':
        return 'tuple'
    elif type1 == 'none' and type2 == 'none':
        return 'none'
    else:
        if type1 == type2:
            raise ValueError("binary operation on type '{}' not supported".format(type1))
        else:
            raise ValueError("type '{}' and type '{}' are incompatible".format(type1, type2))

###########################################################################################
# Asteroid uses truth values similar to Python's Pythonic truth values:
#
# Any object can be tested for truth value, for use in an if or while condition or as
# operand of the Boolean operations.
#
# The following values are considered false:
#
#     none
#     false
#     zero of the numeric types: 0, 0.0.
#     the empty string
#     any empty list: (,), [].
#
#  All other values are considered true, in particular any object is considered
#  to be a true value.
#
def map2boolean(value):

    if value[0] == 'none':
        return ('boolean', False)

    elif value[0] == 'boolean':
        return value

    elif value[0] in  ['integer', 'real', 'list', 'tuple', 'string']:
        return ('boolean', bool(value[1]))

    elif value[0] in ['object', 'function-val', 'pattern']:
        return ('boolean', True)

    else:
        raise ValueError("unsupported type '{}' as truth value".format(value[0]))

###########################################################################################
def data_only(memory):
    '''
    filter an object memory and return a memory with only data values.
    '''
    data_memory = list()

    for item in memory:
        if item[0] != 'function-val':
            data_memory.append(item)

    return data_memory

###########################################################################################
def data_ix_list(memory):
    '''
    compute the set of indexes that point to data members in the object memory
    '''
    ix_list = list()
    for i in range(0,len(memory)):
        if (memory[i])[0] != 'function-val':
            ix_list.append(i)

    return ix_list

###########################################################################################
# the following two tables are used in format processing of strings
# in term2string to support "\\", "\n", "\t", and "\"" characters.
named_re_list = [
    r'(?P<BS2>\\\\)',
    r'(?P<NL>\\n)',
    r'(?P<TAB>\\t)',
    r'(?P<DQ>\\")',
    r'(?P<ANYTHING>.)',
]
combined_re = '|'.join(named_re_list)

def term2string(term):
    TYPE = term[0]

    if TYPE == 'string':
        val = term[1]
        output_str = ""
        match_object_list = list(re.finditer(combined_re, val, re.DOTALL))
        for mo in match_object_list:
            type = mo.lastgroup
            value = mo.group()
            #lhh
            #print(type)
            if type == 'BS2':
                output_str += "\\"
            elif type == 'NL':
                output_str += chr(10)
            elif type == 'TAB':
                output_str += chr(9)
            elif type == 'DQ':
                output_str += '"'
            elif type == 'ANYTHING':
                output_str += value
            else:
                raise ValueError("internal error: unknown match group {} in term2string"
                                 .format(type))
        return output_str

    elif TYPE in ['id', 'integer', 'real']:
        val = term[1]
        return str(val)

    elif TYPE in ['boolean', 'none']:
        val = term[1]
        return str(val).lower()

    elif TYPE in ['list', 'tuple']:
        val = term[1]
        term_string = '[' if TYPE == 'list' else '('
        l = len(val)
        for i in range(l):
            term_string += term2string(val[i])
            if i != l-1:
                term_string += ','
        if l == 1 and TYPE == 'tuple': # proper 1-tuple notation
            term_string += ','
        term_string += ']' if TYPE == 'list' else ')'
        return term_string

    elif TYPE == 'object':
        (OBJECT,
         (STRUCT_ID, (ID, struct_id)),
         (OBJECT_MEMORY, (LIST, object_memory))) = term
        data_memory = data_only(object_memory)
        term_string = struct_id + '('
        for ix in range(0, len(data_memory)):
            term_string += term2string(data_memory[ix])
            term_string += ',' if ix != len(data_memory)-1 else ''
        term_string += ')'
        return term_string

    elif TYPE in ['function-val', 'member-function-val']:
        return '(function ...)'

    elif TYPE == 'apply':
        (APPLY, f, args) = term
        operator_table = {
            '__uminus__'    : ("-", True),
            '__uplus__'     : ("+", True),
            '__not__'       : ("not ", True),
            '__plus__'      : ("+", False),
            '__minus__'     : ("-", False),
            '__times__'     : ("*", False),
            '__divide__'    : ("/", False),
            '__or__'        : (" or ", False),
            '__and__'       : (" and ", False),
            '__eq__'        : (" == ", False),
            '__ne__'        : (" != ", False),
            '__le__'        : (" <= ", False),
            '__lt__'        : (" < ", False),
            '__ge__'        : (" >= ", False),
            '__gt__'        : (" > ", False)
        }
        if isinstance(f[1], str) and f[1] in operator_table:
            (op_symbol, is_unary) = operator_table[f[1]]
            if is_unary:
                term_string = "{}({})".format(
                    op_symbol, term2string(args)
                )
            else:
                (_, arglist) = args
                term_string = "{}{}{}".format(
                    term2string(arglist[0]), op_symbol, term2string(arglist[1])
                )
        else:
            term_string = term2string(f)
            term_string += term2string(args)

        return term_string

    elif TYPE == 'pattern':
        # we are printing out a term - just ignore the pattern operator
        val = term[1]
        return term2string(val)

    elif TYPE == 'nil':
        return ''

    elif TYPE == 'head-tail':           # Handle a head-tail pattern
        length = head_tail_length(term)
        term_string = "["
        for ix in range(1,length):

            #update output text with each entry in head-tail list
            term_string += term2string(term[1])

            #step down the head-tail tree
            term = term[2]

            #Insert head-tail entry delimiter
            term_string+= "|"

            #Catch the last entry
            if (ix == (length-1)):
                term_string += term2string(term)

        #Put the head-tail list delimiter on the end and then return
        term_string += "]"
        return term_string

    elif TYPE == 'raw-head-tail':
        (RHT, e1, e2) = term
        walked_e1 = term2string(e1)
        walked_e2 = term2string(e2)

        return walked_e1 + "|" + walked_e2

    elif TYPE == 'named-pattern':       # Handle a named pattern
        (NAMED_PATTERN,ID,pattern) = term

        if ID[0] == 'index':
            (INDEX, i1, i2) = ID
            term_string = term2string(i1) + "@ " + term2string(i2) + ":"
        else:
            term_string = ID[1] + ':'

        return term_string + term2string(pattern)

    elif TYPE == 'typematch':           # Handle a type pattern
        (TYPECLASS,cType) = term
        term_string = '%'
        term_string += cType
        return term_string

    elif TYPE == 'constraint':          # Handle a constraint-only pattern
        (CONSTRAINT,ptrn) = term
        term_string = '%['
        term_string += term2string(ptrn)
        term_string += ']%'
        return term_string

    elif TYPE == 'deref':               # Handle a first-class pattern
        (DEREF, d_exp) = term
        term_string = "*" + term2string(d_exp)

        #Get the actual pattern from the symbol table
        # NOTE: THIS BREAKS WHEN A PATTERN IS NOT DEFINED IN SCOPE
        # TODO: FIX THIS
        # \/\/\/\/\/\/\/\/\/
        #term_string += term2string(state.symbol_table.lookup_sym(pName))
        return term_string

    elif TYPE == 'if-exp':              # Handle a conditional pattern
        (IF_EXP,EXPRESSION,value,NULL) = term
        # conditions can be arbitrary computations not structural terms
        # therefore we don't print the whole tree.
        return term2string(value)+' if (condition...)'

    elif TYPE == 'index':
        (INDEX, base, ix) = term
        return term2string(base) + "@ " + term2string(ix)

    elif TYPE == 'foreign':
        return "(foreign ...)"

    elif TYPE == 'is':
        (IS, e1, e2) = term

        return term2string(e1) + " is " + term2string(e2)

    elif TYPE == 'in':
        (IN, e1, e2) = term

        return term2string(e1) + " in " + term2string(e2)

    elif TYPE == 'to-list':
        (TO_LIST, (START, start), (STOP, stop), (STRIDE, stride)) = term
        return "[" + term2string(start) + " to " + term2string(stop) + ", stride: " + term2string(stride) + "]"

    elif TYPE == 'function-exp':
        return ('function exp...')

    elif TYPE == 'function-val':
        return ('function val...')

    elif TYPE == 'struct':
        return ('struct...')
    else:
        raise ValueError(
            "unknown type '{}' in term2string"
            .format(TYPE))

##############################################################################################
# Function head_tail_length determines the lenth of a head-tail node by walking to the end.
# The length is then returned from this function as in integer.
# Example Input : [h1|h2|h3|tail]
#         Output: 3
# The output is 3 because the input heal-tail pattern has 3 heads.
#
def head_tail_length( node ):

    # Counter to hold length of head-tail node
    ctr = 1

    # Try to walk down the tree, incrementing the counter
    # return the counter when we fail
    while(1):
        try:
            node = node[2]
        except:
            return ctr
        else:
            ctr += 1

##############################################################################################



###########################################################################################
# Lexer for Asteroid
#
# (c) Lutz Hamel, University of Rhode Island
###########################################################################################

import re

from asteroid.state import state, warning
from asteroid.globals import ExpectationError

# table that specifies the token value and type for keywords
keywords = {
#   value:          type:
    'and'           : 'AND',
    'assert'        : 'ASSERT',
    'break'         : 'BREAK',
    'catch'         : 'CATCH',
    'data'          : 'DATA',
    'do'            : 'DO',
    'elif'          : 'ELIF',
    'else'          : 'ELSE',
    'end'           : 'END',
    'escape'        : 'ESCAPE',
    'eval'          : 'EVAL',
    'for'           : 'FOR',
    'from'          : 'FROM',
    'function'      : 'FUNCTION',
    'global'        : 'GLOBAL',
    'if'            : 'IF',
    'in'            : 'IN',
    'is'            : 'IS',
    'lambda'        : 'LAMBDA',
    'let'           : 'LET',
    'load'          : 'LOAD',
    'loop'          : 'LOOP',
    'not'           : 'NOT',
    'or'            : 'OR',
    'orwith'        : 'ORWITH',
    'pattern'       : 'PATTERN',
    'repeat'        : 'REPEAT',
    'return'        : 'RETURN',
    'step'        : 'STEP',
    'structure'     : 'STRUCTURE',
    'system'        : 'SYSTEM',
    'throw'         : 'THROW',
    'to'            : 'TO',
    'try'           : 'TRY',
    'until'         : 'UNTIL',
    'while'         : 'WHILE',
    'with'          : 'WITH',
    # constants
    'none'          : 'NONE',
    'true'          : 'TRUE',
    'false'         : 'FALSE',
    }

# this table defines tokens whose value is defined by a
# regular expression.
token_specs = [
#   value:                                        display:      type:
    (r'[0-9]+([.][0-9]+)?((e|E)(\+|\-)?[0-9]+)?', "number",     'NUMBER'),
    # the following is very fragil, re's are tried in order, therefore we
    # try matching things in the following order:
    #   1. everything that is not a double quote or a backslash
    #   2. an escaped double quote
    #   3 then a backslash
    (r'"([^"\\]|\\"|\\)*"',                       "string",     'STRING'),
    (r'(--.*)|(\#.*)',                            "comment",    'COMMENT'),
    (r'[a-zA-Z_][a-zA-Z_0-9]*',                   "variable",   'ID'),
    (r'\n',                                       "\n",         'NEWLINE'),
    (r'[ \t]+',                                   "space/tab",  'WHITESPACE'),
    (r'\%[a-zA-Z_][a-zA-Z_0-9]*',                 "%type",      'TYPEMATCH'),
    (r'\+',                                       "'+'",        'PLUS'),
    (r'-',                                        "'-'",        'MINUS'),
    (r'\*',                                       "'*'",        'TIMES'),
    (r'/',                                        "'/'",        'DIVIDE'),
    (r'==',                                       "'=='",       'EQ'),
    (r'=/=',                                      "'=/='",      'NE'),
    (r'<=',                                       "'<='",       'LE'),
    (r'<',                                        "'<'",        'LT'),
    (r'>=',                                       "'>='",       'GE'),
    (r'>',                                        "'>'",        'GT'),
    (r'@',                                        "'@'",        'AT'),
    (r'\%\[',                                     "'%['",       'LCONSTRAINT'),
    (r'\]\%',                                     "']%'",       'RCONSTRAINT'),
    (r'\(',                                       "'('",        'LPAREN'),
    (r'\)',                                       "')'",        'RPAREN'),
    (r'\[',                                       "'['",        'LBRACKET'),
    (r'\]',                                       "']'",        'RBRACKET'),
    (r':',                                        "':'",        'COLON'),
    (r'\|',                                       "'|'",        'BAR'),
    (r'\.',                                       "'.'",        'DOT'),
    (r',',                                        "','",        'COMMA'),
    (r'=',                                        "'='",        'ASSIGN'),
    # this is the catch-all pattern, it has to be
    # here do that we can report illegal characters
    # in the input.
    (r'.',                                        "unknown",    'MISMATCH'),
]

# this table specifies token types that are used in the tokenizer
# but are not defined in the above tables.
implicit_token_types = [
    'INTEGER',
    'REAL',
]

# A table that given a token type will return the expected value
# a reverse token specification: given a type compute the value
token_values = {}

def init_token_values():
    global token_values

    # grab tokens from token_specs table
    token_values.update({'INTEGER':'integer value'})
    token_values.update({'REAL':'real value'})
    token_values.update({'EOF':'EOF'})
    for (_,display,type) in token_specs:
        token_values.update({type:display})

    # grab tokens from keywords table
    for k in keywords:
        token_values.update({keywords[k]:"'"+k+"'"})

def token_lookup(type):
    return token_values[type]

class Token:
    def __init__(self,type,value,module,lineno):
        self.type = type
        self.value = value
        self.module = module
        self.lineno = lineno

    def __str__(self):
        return '({},{},{},{})'.format(self.type,self.value,self.module,self.lineno)

def tokenize(code):
    # output token list
    tokens = []
    # state/line info
    (module, line_num) = state.lineinfo
    # here we create a list of named patterns from the token_specs table
    # the name of the pattern is the token type
    named_re_list = ['(?P<{}>{})'.format(type,re) for (re,_,type) in token_specs]
    # create one giant re that describes the token structure of the whole
    # language. we 'or' together all the re's on the named_re_list
    combined_re = '|'.join(named_re_list)
    # generate a list of match objects. the group name of a match
    # is the token type.
    match_object_list = list(re.finditer(combined_re, code))
    for mo in match_object_list:
        # get the token type and value from
        # the match object
        type = mo.lastgroup
        value = mo.group()
        # some special processing of tokens
        if type == 'NUMBER':
            if '.' in value or 'e' in value or 'E' in value:
                type = 'REAL'
                value = float(value)
            else:
                type = 'INTEGER'
                value = int(value)
        elif type == 'ID':
            # IDs and keywords share the same
            # here we replace the ID type with
            # the appropriate token type given
            # keyword value, if not a keyword the
            # code defaults to the ID token type
            type = keywords.get(value,'ID')
        elif type == 'TYPEMATCH':
            if value[1:] == 'if':
                warning("'%if' has been deprecated, please replace with 'if'")
                type = 'IF'
                value = value[1:]
            else:
                type = 'TYPEMATCH'
                value = value[1:]
        elif type == 'STRING':
            #lhh
            #print(value)
            if value[-2] == '\\' and value[-3] != '\\':
                raise ValueError("bad escape sequence '\\\"' in string")
            lines = value.count('\n')
            (module, lineno) = state.lineinfo
            line_num += lines
            state.lineinfo = (module, line_num)
            value = value[1:-1] # strip the quotes
        elif type == 'NEWLINE':
            line_num += 1
            state.lineinfo = (module,line_num)
            continue
        elif type == 'COMMENT':
            continue
        elif type == 'WHITESPACE':
            continue
        elif type == 'MISMATCH':
            raise ValueError("unexpected character '{}'".format(value))
        # put the token onto the tokens list
        tokens.append(Token(type, value, module, line_num))
    # always append an EOF token so we never run out of tokens
    # in the lexer.
    tokens.append(Token('EOF', '', module, line_num))
    return tokens

def dbg_print(string):
    #print(string)
    pass

# convenient interface to the token stream
class Lexer:
    def __init__(self):
        self.tokens = None
        self.curr_token_ix = None
        self.curr_token = None
        # keep a set of all possible token types in our lexer
        # this let's us weed out bad match calls very easily
        self.token_types = \
            set(type for (_,_,type) in token_specs) | \
            set(list(keywords.values())) | \
            set(implicit_token_types)
        # init token value lookup
        init_token_values()

    def input(self, input_string):
        self.tokens = tokenize(input_string)
        # the following is always valid because we will always have
        # at least the EOF token on the tokens list.
        self.curr_token_ix = 0
        self.curr_token = self.tokens[self.curr_token_ix]
        state.lineinfo = (self.curr_token.module, self.curr_token.lineno)

    def peek(self):
        return self.curr_token

    def next(self):
        dbg_print('skipping {}'.format(self.curr_token.type))
        if self.curr_token.type != 'EOF':
            self.curr_token_ix += 1
            self.curr_token = self.tokens[self.curr_token_ix]
            state.lineinfo = (self.curr_token.module, self.curr_token.lineno)


    def EOF(self):
        if self.curr_token.type == 'EOF':
            return True
        else:
            return False

    def match(self, token_type):
        if token_type not in self.token_types:
            raise ValueError("unknown token type '{}'".format(token_type))
        elif token_type != self.curr_token.type:
            raise ExpectationError(found=token_lookup(self.curr_token.type),
                                   expected=token_lookup(token_type))
        else:
            dbg_print('matching {}'.format(token_type))
            ct = self.curr_token
            self.next()
            return ct

    def match_optional(self, token_type):
        if token_type == self.curr_token.type:
            return self.match(token_type)
        else:
            return None

# test lexer
if __name__ == "__main__":
    from sys import stdin
    lexer = Lexer()
    lexer.input(stdin.read())
    while not lexer.EOF():
        tok = lexer.peek()
        print(tok)
        lexer.next()



###########################################################################################
# front end for Asteroid
#
# (c) Lutz Hamel, University of Rhode Island
###########################################################################################

import os
import sys
from pathlib import Path, PurePath

from asteroid.globals import asteroid_file_suffix, ExpectationError
from asteroid.lex import Lexer, token_lookup
from asteroid.state import state, warning

###########################################################################################
def dbg_print(string):
    #print(string)
    pass

###########################################################################################
# LL(1) lookahead sets

ops = {
    'TIMES',
    'NOT',
    'MINUS',
    'PLUS'
    }

primary_lookahead = {
    'ESCAPE',
    'EVAL',
    'LAMBDA',
    'INTEGER',
    'REAL',
    'STRING',
    'TRUE',
    'FALSE',
    'NONE',
    'ID',
    'LBRACKET',
    'LPAREN',
    'TYPEMATCH',
    } | ops

exp_lookahead = {
    'PATTERN',
    'LCONSTRAINT',} | primary_lookahead

exp_lookahead_no_ops = exp_lookahead - ops

primary_lookahead_no_ops = exp_lookahead_no_ops

stmt_lookahead = {
    'DOT',
    'ASSERT',
    'BREAK',
    'FOR',
    'FUNCTION',
    'GLOBAL',
    'IF',
    'LET',
    'LOAD',
    'LOOP',
    'REPEAT',
    'RETURN',
    'STRUCTURE',
    'THROW',
    'TRY',
    'WHILE',
    #'WITH',
    } | exp_lookahead

###########################################################################################
class Parser:

    def __init__(self, filename="<input>", functional_mode=False):
        self.lexer = Lexer()
        self.functional_mode = functional_mode
        self.system_modules = os.listdir(os.path.dirname( __file__ )+'/modules')
        self.parser_file_path = os.path.split(os.path.dirname(__file__))[0]
        state.lineinfo = (filename,1)

    ###########################################################################################
    # best guess whether we are looking at a system module or not: if the parser
    # file path is a substring of the module file path then the module is a
    # system module
    def is_system_module(self):
        name = state.lineinfo[0]
        if (os.path.split(name)[1] in self.system_modules) and (self.parser_file_path in name):
            return True
        else:
            return False

    ###########################################################################################
    def parse(self, input):
        self.lexer.input(input)
        return self.prog()

    ###########################################################################################
    # prog:
    #   stmt_list
    def prog(self):
        dbg_print("parsing PROG")
        sl = self.stmt_list()
        if not self.lexer.EOF():
            raise SyntaxError("expected 'EOF' found {}." \
                              .format(token_lookup(self.lexer.peek().type)))
        else:
            dbg_print("parsing EOF")
        return sl

    ###########################################################################################
    # stmt_list
    #   : stmt*
    def stmt_list(self):
        dbg_print("parsing STMT_LIST")

        sl = []
        while self.stmt_coming_up():
            sl += [('lineinfo', state.lineinfo)]

            # Get the statment
            stmt = self.stmt()

            # This set of conditionals gives us the behavior we want
            # wrt implicit return values. There's only one situation
            # in which we want to set the return value. That is when
            # a top level expression is the last in a stmt_list

            # If there's a statement coming up or the current statement
            # is not a top level expression, just append the statement
            if self.stmt_coming_up() or stmt[0] != 'top-level-exp':
                sl += [stmt]

            # Otherwhise, if there's no statement coming up and the
            # last statement was a top level expression, set the
            # ret val
            else:
                sl += [('clear-ret-val',)]
                sl += [('set-ret-val', stmt)]             

        return ('list', sl)

    ###########################################################################################
    # NOTE: periods are optional at end of sentences but leaving them out can
    #       lead to ambiguities
    # NOTE: the dot is also short hand for the 'noop' command
    #
    # stmt
    #    : '.' // NOOP
    #    | LOAD SYSTEM? (STRING | ID) '.'?
    #    | GLOBAL id_list '.'?
    #    | ASSERT exp '.'?
    #    | STRUCTURE ID WITH struct_stmts END
    #    | LET pattern '=' exp '.'?
    #    | LOOP DO? stmt_list END
    #    | FOR pattern IN exp DO stmt_list END
    #    | WHILE exp DO stmt_list END
    #    | REPEAT (DO?) stmt_list UNTIL exp '.'?
    #    | BREAK
    #    | IF exp DO stmt_list (ELIF exp DO stmt_list)* (ELSE (DO?) stmt_list)? END
    #    | RETURN exp? '.'?
    #    | TRY stmt_list (CATCH pattern DO stmt_list)+ END
    #    | THROW exp '.'?
    #    | function_def
    #    | exp '.'?
    def stmt(self):
        dbg_print("parsing STMT")
        tt = self.lexer.peek().type  # tt - Token Type

        if tt == 'DOT':
            self.lexer.match('DOT')
            return ('noop',)

        elif tt == 'LOAD':
            # expand the AST from the file into our current AST
            # using a nested parser object
            self.lexer.match('LOAD')
            sys_flag = bool(self.lexer.match_optional('SYSTEM'))
            # allow module names without quotes
            if self.lexer.peek().type in ['STRING', 'ID']:
                str_tok = self.lexer.match(self.lexer.peek().type)
            elif self.lexer.peek().type == 'EOF':
                raise ExpectationError(expected="valid module name", found='EOF' )
            else:
                raise SyntaxError("invalid module name '{}'"
                                  .format(self.lexer.peek().value))
            self.lexer.match_optional('DOT')

            raw_pp = PurePath(str_tok.value)
            module_name = raw_pp.stem

            # if module is on the list of modules then we have loaded
            # it already -- ignore -- continue parsing the program file
            if module_name in state.modules:
                # lhh
                # print("Ignoring module {}".format(module_name))
                return self.stmt_list()

            # search for module file:
            # 0. raw module name - could be an absolute path
            # 1. search in current working directory
            # 2. search in directory where Asteroid is installed
            # TODO: does this work on all OS's?
            # TODO: should have an env variable to set search path
            search_list = []
            if not sys_flag:
                search_list.append(str_tok.value)
                search_list.append(str_tok.value + asteroid_file_suffix)
                search_list.append(os.path.join(os.getcwd(), module_name))
                search_list.append(os.path.join(os.getcwd(), module_name + asteroid_file_suffix))
                search_list.append(os.path.join(os.getcwd(), 'modules', module_name))
                search_list.append(os.path.join(os.getcwd(), 'modules', module_name + asteroid_file_suffix))
            search_list.append(os.path.join(os.path.split(os.path.abspath(__file__))[0], 'modules', module_name))
            search_list.append(os.path.join(os.path.split(os.path.abspath(__file__))[0], 'modules', module_name + asteroid_file_suffix))

            file_found = False

            for ix in range(len(search_list)):
                ast_module_file = search_list[ix]
                #lhh
                #print("AST module: {}".format(ast_module_file))
                # compute Path object so we can test if the file exists
                ast_module_path = Path(ast_module_file)
                if ast_module_path.is_file():
                    file_found = True
                    break

            if not file_found:
                raise ValueError("Asteroid module '{}' not found"
                                 .format(str_tok.value))

            #lhh
            #print("opening module {}".format(ast_module_file))

            old_lineinfo = state.lineinfo
            with open(ast_module_file) as f:
                #state.modules.append(module_name)
                state.modules.append(ast_module_file)
                data = f.read()

                # Give the absolute path to the parser
                fparser = Parser(str(ast_module_path))
                (STMT_LIST, fstmts) = fparser.parse(data)
            
            state.lineinfo = old_lineinfo
            return ('import_stmt', fstmts)

        elif tt == 'GLOBAL':
            dbg_print("parsing GLOBAL")
            self.lexer.match('GLOBAL')
            id_list = self.id_list()
            self.lexer.match_optional('DOT')
            return ('global', id_list)

        elif tt == 'ASSERT':
            dbg_print("parsing ASSERT")
            self.lexer.match('ASSERT')
            exp = self.exp()
            self.lexer.match_optional('DOT')
            return ('assert', exp)

        elif tt == 'FUNCTION':
            return self.function_def()

        elif tt == 'STRUCTURE':
            dbg_print("parsing STRUCTURE")
            self.lexer.match('STRUCTURE')
            id_tok = self.lexer.match('ID')
            self.lexer.match('WITH')
            stmts = self.struct_stmts()
            self.lexer.match('END')
            return ('struct-def',
                    ('id', id_tok.value),
                    ('member-list', stmts))

        elif tt == 'LET':
            dbg_print("parsing LET")
            self.lexer.match('LET')
            p = self.pattern()
            self.lexer.match('ASSIGN')
            v = self.exp()
            self.lexer.match_optional('DOT')
            return ('unify', p, v)

        elif tt == 'LOOP':
            if self.functional_mode and not self.is_system_module():
                raise SyntaxError("loop is not supported in functional mode")
            dbg_print("parsing LOOP")
            self.lexer.match('LOOP')
            self.lexer.match_optional('DO')
            sl = self.stmt_list()
            self.lexer.match('END')
            return ('loop',
                    ('stmt-list', sl))

        elif tt == 'FOR':
            if self.functional_mode and not self.is_system_module():
                raise SyntaxError("for loop is not supported in functional mode")
            dbg_print("parsing FOR")
            self.lexer.match('FOR')
            e = self.exp()
            if e[0] != 'in':
                raise ExpectationError(
                        expected="'in' expression in for loop",
                        found=token_lookup(self.lexer.peek().type))

            self.lexer.match('DO')
            sl = self.stmt_list()
            self.lexer.match('END')
            return ('for',
                    ('in-exp', e),
                    ('stmt-list', sl))

        elif tt == 'WHILE':
            if self.functional_mode and not self.is_system_module():
                raise SyntaxError("while loop is not supported in functional mode")
            dbg_print("parsing WHILE")
            self.lexer.match('WHILE')
            e = self.exp()
            self.lexer.match('DO')
            sl = self.stmt_list()
            self.lexer.match('END')
            #self.lexer.match_optional('WHILE')
            return ('while',
                    ('cond-exp', e),
                    ('stmt-list', sl))

        elif tt == 'REPEAT':
            if self.functional_mode and not self.is_system_module():
                raise SyntaxError("repeat loop is not supported in functional mode")
            dbg_print("parsing REPEAT")
            self.lexer.match('REPEAT')
            self.lexer.match_optional('DO')
            sl = self.stmt_list()
            self.lexer.match('UNTIL')
            e = self.exp()
            self.lexer.match_optional('DOT')
            return ('repeat',
                    ('stmt-list', sl),
                    ('until-exp', e))

        elif tt == 'BREAK':
            if self.functional_mode and not self.is_system_module():
                raise SyntaxError("break statement is not supported in functional mode")
            dbg_print("parsing BREAK")
            self.lexer.match('BREAK')
            return ('break',)

        elif tt == 'IF':
            if self.functional_mode and not self.is_system_module():
                raise SyntaxError("if statement is not supported in functional mode")

            # if statements are coded as a list of ('if-clause', condition, stmts)
            if_list = []

            dbg_print("parsing IF")
            old_lineinfo = state.lineinfo
            self.lexer.match('IF')
            cond = self.exp()
            self.lexer.match('DO')
            stmts = self.stmt_list()
            if_list.append(('lineinfo',old_lineinfo))
            if_list.append(('if-clause', ('cond', cond), ('stmt-list', stmts)))

            while self.lexer.peek().type == 'ELIF':
                dbg_print("parsing ELIF")
                old_lineinfo = state.lineinfo
                self.lexer.match('ELIF')
                cond = self.exp()
                self.lexer.match('DO')
                stmts = self.stmt_list()
                if_list.append(('lineinfo',old_lineinfo))
                if_list.append(('if-clause', ('cond', cond), ('stmt-list', stmts)))

            if self.lexer.peek().type == 'ELSE':
                dbg_print("parsing ELSE")
                old_lineinfo = state.lineinfo
                self.lexer.match('ELSE')
                self.lexer.match_optional('DO')
                stmts = self.stmt_list()
                if_list.append(('lineinfo',old_lineinfo))
                # make the else look like another elif with the condition set to 'true'
                if_list.append(('if-clause', ('cond', ('boolean', True)), ('stmt-list', stmts)))

            self.lexer.match('END')
            return ('if', ('list', if_list))


        elif tt == 'RETURN':
            dbg_print("parsing RETURN")
            self.lexer.match('RETURN')
            if self.lexer.peek().type in exp_lookahead:
                e = self.exp()
                self.lexer.match_optional('DOT')
                return ('return', e)
            else:
                self.lexer.match_optional('DOT')
                return ('return', ('none', None))

        elif tt == 'TRY':
            dbg_print("parsing TRY")

            # the catch list is a list of ('catch', pattern, stmts)
            catch_list = []

            self.lexer.match('TRY')
            try_stmts = self.stmt_list()
            self.lexer.match('CATCH')
            dbg_print("parsing CATCH")
            pattern = self.pattern()
            self.lexer.match('DO')
            stmts = self.stmt_list()
            catch_list.append(('catch', ('pattern', pattern), ('stmt-list', stmts)))

            while self.lexer.peek().type == 'CATCH':
                dbg_print("parsing CATCH")
                self.lexer.match('CATCH')
                pattern = self.pattern()
                self.lexer.match('DO')
                stmts = self.stmt_list()
                catch_list.append(('catch',('pattern', pattern), ('stmt-list', stmts)))

            self.lexer.match('END')
            #self.lexer.match_optional('TRY')

            return ('try',
                    ('stmt-list', try_stmts),
                    ('catch-list', ('list', catch_list)))

        elif tt == 'THROW':
            dbg_print("parsing THROW")
            self.lexer.match('THROW')
            e = self.exp()
            self.lexer.match_optional('DOT')
            return ('throw', e)

        elif tt in exp_lookahead:
            v = self.exp()
            self.lexer.match_optional('DOT')
            return ('top-level-exp', v)

        else:
            raise SyntaxError("syntax error at '{}'"
                        .format(self.lexer.peek().value))

    ###########################################################################################
    # function_def
    #  : FUNCTION ID body_defs END FUNCTION
    def function_def(self):
        dbg_print("parsing FUNCTION_DEF")
        self.lexer.match('FUNCTION')
        id_tok = self.lexer.match('ID')
        body_list = self.body_defs()
        self.lexer.match('END')
        #self.lexer.match('FUNCTION')

        # check if any useless patterns exist within the function
        #lhh check_redundancy( body_list, id_tok )

        # functions are function expressions bound to names
        return ('unify',
                ('id',id_tok.value),
                ('function-exp', body_list))

    ###########################################################################################
    # data_stmt
    #  : DATA ID
    def data_stmt(self):
        dbg_print("parsing DATA_STMT")

        if self.lexer.peek().type == 'DATA':
            self.lexer.match('DATA')
            id_tok = self.lexer.match('ID')
            return ('data', ('id', id_tok.value))
        else:
            raise SyntaxError(
                "syntax error at '{}'"
                .format(self.lexer.peek().value))

    ###########################################################################################
    # struct_stmt
    #   : data_stmt '.'?
    #   | function_def '.'?
    #   | '.'
    def struct_stmt(self):
        dbg_print("parsing STRUCT_STMT")

        if self.lexer.peek().type == 'DATA':
            s = self.data_stmt()
            self.lexer.match_optional('DOT')
            return s
        elif self.lexer.peek().type == 'FUNCTION':
            s = self.function_def()
            self.lexer.match_optional('DOT')
            return s
        elif self.lexer.peek().type == 'DOT':
            self.lexer.match('DOT')
            return ('noop',)
        else:
            raise SyntaxError(
                "syntax error at '{}'"
                .format(self.lexer.peek().value))
    ###########################################################################################
    # struct_stmts
    #   : struct_stmt*
    def struct_stmts(self):
        dbg_print("parsing STRUCT_STMTS")

        sl = []
        while self.lexer.peek().type in ['DATA', 'FUNCTION', 'DOT']:
            sl += [self.struct_stmt()]
        return ('list', sl)

    ###########################################################################################
    # id_list
    #   : ID (',' ID)*
    def id_list(self):
        dbg_print("parsing ID_LIST")

        id_list = []

        id_tok = self.lexer.match('ID')
        id_list.append(('id', id_tok.value))
        while self.lexer.peek().type == 'COMMA':
            self.lexer.match('COMMA')
            id_tok = self.lexer.match('ID')
            id_list.append(('id', id_tok.value))
        return ('list', id_list)

    ###########################################################################################
    # body_defs
    #   : WITH pattern DO stmt_list (ORWITH pattern DO stmt_list)*
    def body_defs(self):
        dbg_print("parsing BODY_DEFS")

        # a list of ('body', pattern, stmts) pairs
        body_list = []

        cur_lineinfo = state.lineinfo
        self.lexer.match('WITH')
        p = self.pattern()
        self.lexer.match('DO')
        sl = self.stmt_list()
        body_list.append( ('lineinfo', cur_lineinfo) )
        body_list.append(('body', ('pattern', p), ('stmt-list', sl)))

        while self.lexer.peek().type in ['ORWITH','WITH']:
            cur_lineinfo = state.lineinfo

            if self.lexer.peek().type == 'ORWITH':
                warning("'orwith' has been deprecated, please replace with 'with'")
            self.lexer.match(self.lexer.peek().type)
            p = self.pattern()
            self.lexer.match('DO')
            sl = self.stmt_list()
            body_list.append( ('lineinfo', cur_lineinfo) )
            body_list.append(('body', ('pattern', p), ('stmt-list', sl)))

        return ('body-list', ('list', body_list))

    ###########################################################################################
    # exp
    #    : pattern
    def exp(self):
        dbg_print("parsing EXP")
        v = self.pattern()
        return v

    ###########################################################################################
    # pattern
    #    : PATTERN WITH? exp
    #    | '%[' exp ']%'
    #    | head_tail
    def pattern(self):
        dbg_print("parsing PATTERN")

        if self.lexer.peek().type == 'PATTERN':
            self.lexer.match('PATTERN')
            self.lexer.match_optional('WITH')
            v = self.exp()
            return ('pattern', v)
        elif self.lexer.peek().type == 'LCONSTRAINT': #constraint-only pattern match
            self.lexer.match('LCONSTRAINT')
            v = self.exp()
            self.lexer.match('RCONSTRAINT')
            return ('constraint', v)
        else:
            v = self.head_tail()
            return v

    ###########################################################################################
    # head_tail
    #    : conditional ('|' exp)?
    #
    # NOTE: * as a value this operator will construct a list from the semantic values of
    #         head and tail
    #       * as a pattern this operator will be unified with a list such that head will
    #         unify with the first element of the list and tail with the remaining list
    # NOTE: this is a list constructor and therefore should never appear in the semantic
    #       processing, use walk to expand the list before processing it.
    def head_tail(self):
        dbg_print("parsing HEAD_TAIL")

        v = self.conditional()

        if self.lexer.peek().type == 'BAR':
            self.lexer.match('BAR')
            head = v
            tail = self.exp()
            v = ('raw-head-tail', head, tail)

        return v

    ###########################################################################################
    # conditional patterns are now supported via 'pattern if cond'
    # no else part. Since this overlaps with conditional expressions
    # we check for correct usage semantically.
    #
    # conditional
    #    : compound (IF exp (ELSE exp)?)?
    def conditional(self):
        dbg_print("parsing CONDITIONAL")

        v = self.compound()

        tt = self.lexer.peek().type

        if tt == 'IF':
            self.lexer.match('IF')
            v2 = self.exp()
            if self.lexer.peek().type == 'ELSE':
                self.lexer.match('ELSE')
                v3 = self.exp()
            else:
                # this happens when the conditional is used in a
                # conditional pattern
                v3 = ('null',)
            return ('if-exp', v2, v, v3) # mapping it into standard if-then-else format

        else:
            return v

    ###########################################################################################
    # compound
    #    : logic_exp0
    #        (
    #           (IS pattern) |
    #           (IN exp) | // exp has to be a list
    #           (TO exp (STEP exp)?) | // list comprehension
    #        )?
    def compound(self):
        dbg_print("parsing COMPOUND")

        v = self.logic_exp0()

        tt = self.lexer.peek().type
        if tt == 'IS':
            self.lexer.match('IS')
            v2 = self.pattern()
            return ('is', v, v2)

        elif tt == 'IN':
            self.lexer.match('IN')
            v2 = self.exp()
            return ('in', v, v2)

        elif tt == 'TO':
            self.lexer.match('TO')
            v2 = self.exp()
            if self.lexer.peek().type == 'STEP':
                self.lexer.match('STEP')
                v3 = self.exp()
                return ('raw-to-list',
                        ('start', v),
                        ('stop', v2),
                        ('step', v3))
            else:
                return ('raw-to-list',
                        ('start', v),
                        ('stop', v2),
                        ('step', ('integer', '1')))

        else:
            return v

    ###########################################################################################
    # NOTE: Builtin operators are mapped to 'apply' so that they don't have to be
    #       special cased during pattern matching.  See operator_symbols above.
    ###########################################################################################
    # logic/relational/arithmetic operators with their precedence
    # logic_exp0
    #   : logic_exp1 (OR logic_exp1)*
    #
    # logic_exp1
    #   : rel_exp1 (AND rel_exp1)*
    #
    # rel_exp0
    #   : rel_exp1 (('==' | '=/=' /* not equal */) rel_exp1)*
    #
    # rel_exp1
    #   : arith_exp0 (('<=' | '<'  | '>=' | '>') arith_exp0)*
    #
    # arith_exp0
    #   : arith_exp1 (('+' | '-') arith_exp1)*
    #
    # arith_exp1
    #   : call_or_index (('*' | '/') call_or_index)*
    #
    def logic_exp0(self):
        dbg_print("parsing LOGIC/REL/ARITH EXP")
        v = self.logic_exp1()
        while self.lexer.peek().type == 'OR':
            self.lexer.match('OR')
            v2 = self.logic_exp1()
            op_sym = '__or__'
            v = ('apply', ('id', op_sym), ('tuple', [v, v2]))
        return v

    def logic_exp1(self):
        v = self.rel_exp0()
        while self.lexer.peek().type == 'AND':
            self.lexer.match('AND')
            v2 = self.rel_exp0()
            op_sym = '__and__'
            v = ('apply', ('id', op_sym), ('tuple', [v, v2]))
        return v

    def rel_exp0(self):
        v = self.rel_exp1()
        while self.lexer.peek().type in ['EQ', 'NE']:
            op_tok = self.lexer.peek()
            self.lexer.next()
            v2 = self.rel_exp1()
            op_sym = '__' + op_tok.type.lower() + '__'
            v = ('apply', ('id', op_sym), ('tuple', [v, v2]))
        return v

    def rel_exp1(self):
        v = self.arith_exp0()
        while self.lexer.peek().type in ['LE', 'LT', 'GE', 'GT']:
            op_tok = self.lexer.peek()
            self.lexer.next()
            v2 = self.arith_exp0()
            op_sym = '__' + op_tok.type.lower() + '__'
            v = ('apply', ('id', op_sym), ('tuple', [v, v2]))
        return v

    def arith_exp0(self):
        dbg_print("parsing ARITH_EXP")
        v = self.arith_exp1()
        while self.lexer.peek().type in ['PLUS', 'MINUS']:
            op_tok = self.lexer.peek()
            self.lexer.next()
            v2 = self.arith_exp1()
            op_sym = '__' + op_tok.type.lower() + '__'
            v = ('apply', ('id', op_sym), ('tuple', [v, v2]))
        return v

    def arith_exp1(self):
        v = self.call_or_index()
        while self.lexer.peek().type in ['TIMES', 'DIVIDE']:
            op_tok = self.lexer.peek()
            self.lexer.next()
            v2 = self.call_or_index()
            op_sym = '__' + op_tok.type.lower() + '__'
            v = ('apply', ('id', op_sym), ('tuple', [v, v2]))
        return v

    ###########################################################################################
    # call_or_index
    #   : primary (primary | '@' primary)* (':' pattern)?  // named pattern when ': pattern' exists
    def call_or_index(self):
        dbg_print("parsing CALL_OR_INDEX")

        v = self.primary()

        # Note: the 'no ops' lookahead here is necessary because operators
        # can never be arguments to a function in Asteroid
        call_or_index_lookahead = primary_lookahead_no_ops|set(['AT'])
        while self.lexer.peek().type in call_or_index_lookahead:
            if self.lexer.peek().type in primary_lookahead:
                v2 = self.primary()
                v = ('apply', v, v2)
            elif self.lexer.peek().type == 'AT':
                self.lexer.match('AT')
                ix = self.primary()
                v = ('index', v, ix)

        if self.lexer.peek().type == 'COLON': # if ':' exists - named pattern
            self.lexer.match('COLON')
            e = self.exp()
            v = ('named-pattern', v, e)

        return v

    ###########################################################################################
    # primary
    #    : INTEGER
    #    | REAL
    #    | STRING
    #    | TRUE
    #    | FALSE
    #    | NONE
    #    | ID
    #    | '*' call_or_index /* pattern dereferencing */
    #    | NOT call_or_index
    #    | MINUS call_or_index
    #    | PLUS call_or_index
    #    | ESCAPE STRING
    #    | EVAL exp
    #    | '(' tuple_stuff ')' /* tuple/parenthesized expr */
    #    | '[' list_stuff ']'  /* list or list access */
    #    | function_const
    #    | TYPEMATCH // TYPEMATCH == '%'<typename>
    def primary(self):
        dbg_print("parsing PRIMARY")

        tt = self.lexer.peek().type

        if tt == 'INTEGER':
            tok = self.lexer.match('INTEGER')
            return ('integer', tok.value)

        elif tt == 'REAL':
            tok = self.lexer.match('REAL')
            return ('real', tok.value)

        elif tt == 'STRING':
            tok = self.lexer.match('STRING')
            return ('string', tok.value)

        elif tt == 'TRUE':
            self.lexer.match('TRUE')
            return ('boolean', True)

        elif tt == 'FALSE':
            self.lexer.match('FALSE')
            return ('boolean', False)

        elif tt == 'NONE':
           self.lexer.match('NONE')
           return ('none', None)

        elif tt == 'ID':
            tok = self.lexer.match('ID')
            return ('id', tok.value)

        elif tt == 'TIMES':
            self.lexer.match('TIMES')
            v = self.call_or_index()
            return ('deref', v)

        elif tt == 'NOT':
            self.lexer.match('NOT')
            v = self.call_or_index()
            return ('apply', ('id', '__not__'), v)

        elif tt == 'MINUS':
            self.lexer.match('MINUS')
            v = self.call_or_index()
            # if v is a real or integer constant we apply __uminus__
            if v[0] in ['integer', 'real']:
                return (v[0], - v[1])
            else:
                return ('apply', ('id', '__uminus__'), v)

        elif tt == 'PLUS':
            self.lexer.match('PLUS')
            v = self.call_or_index()
            # if v is a real or integer constant we apply __uplus__
            if v[0] in ['integer', 'real']:
                return (v[0], + v[1])
            else:
                return ('apply', ('id', '__uplus__'), v)

        elif tt == 'ESCAPE':
            self.lexer.match('ESCAPE')
            str_tok = self.lexer.match('STRING')
            return ('escape', str_tok.value)

        elif tt == 'EVAL':
            self.lexer.match('EVAL')
            exp = self.primary()
            return ('eval', exp)

        elif tt == 'LPAREN':
            # Parenthesized expressions have the following meaning:
            #       (A)    means a parenthesized value A
            #       (A,)   means a tuple with a single value A
            #       (A, B) means a tuple with values A and B
            #       ()     shorthand for 'none'
            self.lexer.match('LPAREN')
            v = self.tuple_stuff()
            self.lexer.match('RPAREN')
            return v

        elif tt == 'LBRACKET':
            self.lexer.match('LBRACKET')
            v = self.list_stuff()
            self.lexer.match('RBRACKET')
            return v

        elif tt == 'LAMBDA':
            return self.function_const()

        elif tt == 'TYPEMATCH':
            tok = self.lexer.match('TYPEMATCH')
            return ('typematch', tok.value)

        else:
            raise ExpectationError(
                expected='expression',
                found=token_lookup(self.lexer.peek().type))

            #raise SyntaxError("syntax error at '{}'"
                # .format(self.lexer.peek().value))

    ###########################################################################################
    # tuple_stuff
    #   : exp (',' exp?)*
    #   | empty
    def tuple_stuff(self):
        dbg_print("parsing TUPLE_STUFF")
        if self.lexer.peek().type in exp_lookahead:
            v = self.exp()
            if self.lexer.peek().type == 'COMMA': # if ',' exists - tuple!
                tuple_list = [v]
                while self.lexer.peek().type == 'COMMA':
                    self.lexer.match('COMMA')
                    if self.lexer.peek().type in exp_lookahead:
                        e = self.exp()
                        tuple_list.append(e)
                return ('tuple', tuple_list)

            else: # just parenthesized value - drop parentheses
                return v
        else:
            # empty parentheses are a shorthand for 'none'
            return ('none', None)

    ###########################################################################################
    # list_stuff
    #   : exp (',' exp)*
    #   | empty
    def list_stuff(self):
        dbg_print("parsing LIST_STUFF")
        if self.lexer.peek().type in exp_lookahead:
            v = self.exp()
            if v[0] == 'raw-to-list':
                return ('to-list', v[1], v[2], v[3])
            elif v[0] == 'raw-head-tail':
                return ('head-tail', v[1], v[2])
            elif self.lexer.peek().type == 'COMMA': # if ',' exists - list!
                list_list = [v]
                while self.lexer.peek().type == 'COMMA':
                    self.lexer.match('COMMA')
                    e = self.exp()
                    list_list.append(e)
                return ('list', list_list)
            else:
                return ('list', [v])
        else:
            return ('list', [])

    ###########################################################################################
    # function_const
    #    : LAMBDA body_defs
    def function_const(self):
        dbg_print("parsing FUNCTION_CONST")
        self.lexer.match('LAMBDA')
        body_list = self.body_defs()

        return ('function-exp', body_list)

    ###########################################################################################
    # Minor helper function
    def stmt_coming_up(self):
        return self.lexer.peek().type in stmt_lookahead


###########################################################################################
# Asteroid State Object
#
# (c) University of Rhode Island
###########################################################################################

from asteroid.symtab import SymTab

class State:
    def __init__(self):
        self.initialize()

    def initialize(self,module="<input>"):
        self.symbol_table = SymTab()
        self.modules = [] # loaded modules
        self.AST = None
        self.ignore_pattern = 0 # used to evaluate pattern expressions
        self.constraint_lvl = 0 # used to evaluate constraint-only patterns
        self.cond_warning = False # used to indicate if conditional subsumption
                                  # warning has been displayed
        self.eval_redundancy = True
        self.lineinfo = (module, 1) # tuple: module, lineno
        # stack of 3-tuples for stack trace of function
        # calls: (module,lineno,function name)
        self.trace_stack = [(module,1,"<toplevel>")]

state = State()

def warning(str):
    module, lineno = state.lineinfo
    print("Warning: {}: {}: {}".format(module, lineno, str))

def dump_trace():
    if len(state.trace_stack) == 1:
        return
    else:
        print("traceback (most recent call last):")
        for i in range(0,len(state.trace_stack)):
            (module,lineno,fname) = state.trace_stack[i]
            print("{}: {}: calling {}".format(module,lineno,fname))



###########################################################################################
# Asteroid interpreter
#
# (c) University of Rhode Island
###########################################################################################

import os
import sys

from asteroid.globals import *
from asteroid.support import *
from asteroid.frontend import Parser
from asteroid.state import state, dump_trace
from asteroid.walk import walk_program, debug_walk

# the prologue file is expected to be in the 'modules' folder
prologue_name = 'prologue.ast'
def load_prologue():
    """
    Load the prolog files and run them.
    """
    prologue_file = ''
    prologue_file_base = os.path.join('modules', prologue_name)
    module_path = os.path.join(os.path.split(os.path.abspath(__file__))[0], prologue_file_base)
    working_path = os.path.join(os.getcwd(), prologue_file_base)

    if os.path.isfile(module_path):
        prologue_file = module_path
    
    elif os.path.isfile(working_path):
        prologue_file = working_path
    
    else:
        raise ValueError("Asteroid prologue '{}' not found"
                        .format(prologue_file_base))
    
    with open(prologue_file) as f:
        state.modules.append(prologue_name)
        data = f.read()
        pparser = Parser(prologue_file)
        (LIST, pstmts) = pparser.parse(data)

    state.AST = ('list', pstmts)
    walk_program(state.AST)
    state.AST = None

def interp(program,
           program_name = "<input>",
           tree_dump=False,
           do_walk=True,
           symtab_dump=False,
           exceptions=False,
           redundancy=True,
           prologue=True,
           debugger=None,
           functional_mode=False,
           initialize_state = True
           ):
    '''
    The function 'interp' is the top-level entry point to the
    Asteroid interpreter with the following arguments:
      * program: a string representing an Asteroid program
      * program_name: the of the program to be interpreted
      * tree_dump: a flag indicating whether the AST should be dumped after parsing
      * do_walk: a flag indicating whether the interpret should interpret the AST
      * symtab_dump: a flag indicating whether the contents of the symbol
                     table should be printed out
      * exceptions: when true Python style exceptions are shown, otherwise
                    Asteroid style exceptions are shown
      * redundancy: a flag indicating whether the function pattern checker should be run
      * prologue: a flag indicating whether the Asteroid prologue file should be loaded
      * functional_mode: if set then the Asteroid interpreter behaves like an interpreter
                         functional programming language.
      * initialize_state: if set then the interpreter will (re)initialize its state.  
    '''
    try:
        # initialize state
        if initialize_state:
            state.initialize(program_name)

        if prologue:
            load_prologue()

        # initialize "check for useless clauses" flag
        state.eval_redundancy = redundancy

        # build the AST
        parser = Parser(program_name, functional_mode)
        (LIST, istmts) = parser.parse(program)
        state.AST = ('list', istmts)

        # walk the AST
        if tree_dump:
            dump_AST(state.AST)
        if do_walk:
            if debugger:
                debug_walk(state.AST, debugger)
            else:
                walk_program(state.AST)
        if symtab_dump:
            state.symbol_table.dump()

    # Note: all exceptions generated by Asteroid should go through the handler
    #       for 'Exception' unless the exception needs special handling like
    #       'ThrowValue' or 'ReturnValue' etc.
    except ThrowValue as throw_val:
        dump_trace()
        # handle exceptions using the standard Error constructor
        module, lineno = state.lineinfo
        if throw_val.value[0] == 'apply' and throw_val.value[1][1] == 'Error':
            (APPLY, (ID, id), (APPLY, error_obj, rest)) = throw_val.value
            print("error: {}: {}: {}".format(module, lineno, term2string(error_obj)))
        else:
            print("error: {}: {}: unhandled Asteroid exception: {}"
                  .format(module, lineno, term2string(throw_val.value)))
        # needed for REPL
        if not exceptions:
            sys.exit(1)

    except ReturnValue as inst:
        dump_trace()
        module, lineno = state.lineinfo
        print("error: {}: {}: return statement used outside a function environment"
              .format(module, lineno))
        # needed for REPL
        if not exceptions:
            sys.exit(1)

    except  KeyboardInterrupt as e:
        if debugger:
            raise e

        dump_trace()
        print("error: keyboard interrupt")
        # needed for REPL
        if not exceptions:
            sys.exit(1)

    except SystemExit:
        exit(0)

    except Exception as e:
        if exceptions: # rethrow the exception so that you can see the full backtrace
            if symtab_dump:
                state.symbol_table.dump()
            raise e
        else:
            dump_trace()
            module, lineno = state.lineinfo
            print("error: {}: {}: {}".format(module, lineno, e))
            # needed for REPL
            if not exceptions:
                sys.exit(1)

    except BaseException as e:
        dump_trace()
        print("error: {}".format(e))
        # needed for REPL
        if not exceptions:
            sys.exit(1)


###########################################################################################
# Asteroid interpreter shell
#
# (c) University of Rhode Island
###########################################################################################

from asteroid.interp import interp, load_prologue
from asteroid.version import VERSION
from asteroid.state import state
from asteroid.globals import ExpectationError
from asteroid.walk import function_return_value
from asteroid.support import term2string

from sys import stdin
import platform

if platform.system() == 'Windows':
    import pyreadline3
else:
    import readline

def repl(new=True, redundancy=False, prologue=False, functional_mode=False):

    if new:
        state.initialize()
        load_prologue()
        print_repl_menu()
    try:
        run_repl(redundancy, prologue, functional_mode)
    except EOFError:
        print()
        pass

def print_repl_menu():
    print("Asteroid Version", VERSION)
    print("(c) University of Rhode Island")
    print("Type \"asteroid -h\" for help")
    if platform.system() == 'Windows':
        print("Press CTRL-Z + Return to exit")
    else:
        print("Press CTRL-D to exit")

def run_repl(redundancy, prologue, functional_mode):

    # The two different prompt types either > for a new statement
    # or . for continuing one
    # lhh: changed the prompt since replit.com uses the > as their console prompt
    arrow_prompt, continue_prompt = ("ast> ", ".... ")
    current_prompt = arrow_prompt

    # Our line to be interpreted
    line = ""
    while True:
        """
        Line input, breaking, and exiting
        """
        try:
            # Get the new input and append it to the previous line (Possibly empty)
            # with a newline in between

            # If the line is empty, just set the line
            if line == "":
                line = input(current_prompt)

            # Otherwhise append a new line
            else:
                line += "\n" + input(current_prompt)

        except KeyboardInterrupt:
            line = ""
            current_prompt = arrow_prompt
            print()
            continue

        except EOFError:
            print()
            break

        """
        Interpretation, multiline input, and exception handling
        """
        try:
            # Try to interpret the new statement
            interp(line,
                   initialize_state=False,
                   redundancy=redundancy,
                   prologue=prologue,
                   functional_mode=functional_mode,
                   exceptions=True)

            # Try to
            line = ""

            # Check for return value
            if function_return_value[-1]:
                # Get the last return value (type, value)
                retval = function_return_value[-1]

                # If it isn't none, print out the value
                if retval[1] != None:
                    print(term2string(function_return_value[-1]))

                    # Reset the return value
                    function_return_value[0] = None

        except ExpectationError as e:
            # If we expected something but found EOF, it's a continue
            if e.found_EOF:
                current_prompt = continue_prompt
            else:
                print("error: "+str(e))
                line = ""
                current_prompt = arrow_prompt

        except Exception as e:
            # FIX THIS
            print("error: "+str(e))
            line = ""
            current_prompt = arrow_prompt
        else:
            current_prompt = arrow_prompt



command_description_table = {
    'macro':        """macro (name) = (command list)
    Define a macro with name "name" and a list of commands.

    Typing solely "macro" will list off the macros currently enabled
    Example:
        macro see_x = continue; explicit; n; p x;
    """,
    
    'step':         """[s]tep
    Step to the next executing line
    """,

    'until':        """[u]ntil ?lineno
    Without a given line number, continue execution until a line that is greater than the current
    one is reached.

    With a line number, continue execution until a line that is greater than or equal to that one
    is reached.

    In both cases, stop when the current frame returns.
    """,
    
    'continue':     """[[c]ont]inue
    Continue to the next breakpoint
    """,
    
    'next':         """[n]ext
    Go to the next top level statement
    """,
    
    'break':        """[b]reak (linenum (if eval("command")))
    Set a breakpoint at linenum. If you supply a command, the breakpoint
    becomes conditional
    """,

    'eval' :     """eval("command")
    Execute asteroid code
    Example:
        eval("let x = 10. io@println(\\"10\\")")

    -- Print out the contents of variable x
        eval("x")
    """,
    
    'delete':       """[[d]el]ete (linenum)
    Delete the breakpoint at linenum
    """,
    
    '!':            """!
    Start a REPL session
    """,
    
    'longlist':     """[ll] longlist
    List the contents of the entire program
    """,
    
    'list':         """[l]ist
    List the contents of the program around the current line
    """,

    'retval':       """[[r]et]val
    Print the most recent return value
    """,
    
    '>':            """>
    Move down one frame
    """,

    '<':            """<
    Move up one frame
    """,

    'where':        """[w]here
    Lists the currently available stack frames and shows the user
    the currently active one.
    """,

    'quit':         """[q]uit
    Exit ADB
    """,
    
    'explicit':     """[e]xplicit (on|off)?
    Toggle explicit mode or set explicit mode to a specific state.

    For more information about explicit mode, refer to "ADB in action".

    Example:
        explicit    -- Toggles explicit mode
        explicit on -- Turns explicit mode on
        explcit off -- Turns explicit mode off
    """,
    
    'help':         """[h]elp ?(name)
    Get help for a command
    """
}



#########################################################################
# A tree walker to interpret Asteroid programs
#
# (c) University of Rhode Island
#########################################################################

from copy import deepcopy
from re import match as re_match

from asteroid.globals import *
from asteroid.support import *
from asteroid.state import state, warning

#########################################################################
# These two variables are used by the debugger. "Debugging" is a flag
# to tell if we're currently in debugging mode and "Debugger" is a
# debugger object. For more information on the debugger's helper
# functions. See the function definitions in this file.
#########################################################################
debugging = False
debugger = None

#########################################################################
# this dictionary maps list member function names to function
# implementations given in the Asteroid prologue.
# see 'prologue.ast' for details
list_member_functions = dict()

#########################################################################
# this dictionary maps string member function names to function
# implementations given in the Asteroid prologue.
# see 'prologue.ast' for details
string_member_functions = dict()

#########################################################################
__retval__ = None  # return value register for escaped code

#########################################################################
# return values for function computed by the last expression executed
# in the context of a function.  note that we consider global code
# to be part of the 'top-level' function
function_return_value = [None]

###########################################################################################
def __unify(term, pattern, unifying = True ):
    '''
    unify term and pattern recursively and return the unifier.
    this unification allows for the same variable to appear
    multiple times in the unifier.  the user of this
    function must take appropriate actions if this happens.

    we assume that both the term and the pattern are made up of tuple
    nodes:

             (<type>, children*)

    leaf nodes must be nullary constructors.

    NOTE: if the pattern looks like an lval then it is treated like an lval, e.g.
            let a@[0] = 'a@[0].
          stores the term 'a@[0] into lval a@[0].
    NOTE: Default argument unifying is set to be true. If we are unifying, then we are
          evaluating unification between a pattern and a term. If we are not
          unifying, then we are evaluating subsumption between two patterns for the
          purpose of detecting redundant/useless pattern clauses in functions.
    '''

    ### Python value level matching
    # NOTE: in the first rules where we test instances we are comparing
    # Python level values, if they don't match exactly then we have
    # a pattern match fail.
    if isinstance(term, str): # apply regular expression match
        message_explicit("Regex match: {} and {}", [pattern, term], level="secondary")

        if isinstance(pattern, str) and re_match("^"+pattern+"$", term):
            # Note: a pattern needs to match the whole term.
            message_explicit("Matched!", level="tertiary")
            return [] # return empty unifier
        else:
            message_explicit("Failed!", level="tertiary")
            raise PatternMatchFailed(
                "regular expression '{}' did not match '{}'"
                .format(pattern, term))

    elif isinstance(term, (int, float, bool)):
        message_explicit("Literal match: {} and {}", [pattern, term], level="primary")

        if term == pattern:
            message_explicit("Matched!", level="secondary")
            return [] # return an empty unifier
        else:
            message_explicit("Failed! {} != {}", [pattern, term], level="secondary")
            raise PatternMatchFailed(
                "'{}' is not the same as '{}'"
                .format(term, pattern))

    elif isinstance(term, list) or isinstance(pattern, list):
        if not(isinstance(term, list)) or not(isinstance(pattern, list)):
            raise PatternMatchFailed(
                "term and pattern do not agree on list/tuple constructor")
        elif len(term) != len(pattern):
            raise PatternMatchFailed(
                "term and pattern lists/tuples are not the same length")
        else:
            message_explicit("Matching lists: {} and {}",
                [gen_t2s( ('list', pattern) ), gen_t2s( ('list', term) )],
                increase=True, notify=True)

            # Make our unifier(s)
            unifier = []
            for i in range(len(term)):
                unifier += unify(term[i], pattern[i], unifying)

            message_explicit("Matched!", decrease=True)

            # Ensure we have no non-linear patterns
            check_repeated_symbols(unifier)

            return unifier
    
    elif ((not unifying) and (term[0] == 'named-pattern')):

        # Unpack a term-side name-pattern if evaluating redundant clauses
        message_explicit("Evaluating named pattern", notify=True)
        return unify(term[2],pattern,unifying)

    elif ((not unifying) and (term[0] == 'deref')):

        # Unpack a term-side first-class pattern if evaluating redundant clauses
        term_pattern = walk(term[1])
        return unify(term_pattern, pattern, unifying)

    ### Asteroid value level matching
    elif pattern[0] == 'object' and term[0] == 'object':
        # this can happen when we dereference a variable pointing
        # to an object as a pattern, e.g.
        #    let o = A(1,2). -- A is a structure with 2 data members
        #    let *o = o.
        (OBJECT, (STRUCT_ID, (ID, pid)), (OBJECT_MEMORY, (LIST, pl))) = pattern
        (OBJECT, (STRUCT_ID, (ID, tid)), (OBJECT_MEMORY, (LIST, tl))) = term
        if pid != tid:
            raise PatternMatchFailed(
                "pattern type '{}' and term type '{}' do not agree"
                .format(pid,tid))

        message_explicit("Matching objects {}{} and {}{}",
            [pid, gen_t2s( ('tuple', pl) ), tid, gen_t2s( ('tuple', tl) )],
            increase=True
        )

        unifiers = []

        for i in range(len(pl)):
            # OWM: We can't actually unify function-vals. If we know the type is
            # the same, can we just assume the function-val is the same?
            notify_explicit()

            if tl[i][0] != 'function-val':
                unifiers += unify(tl[i], pl[i])
        
        message_explicit("Objects matched", decrease=True)
        return unifiers

    elif pattern[0] == 'string' and term[0] != 'string':
        # regular expression applied to a non-string structure
        # this is possible because all data types are subtypes of string
        message_explicit("Matching string {} and non-string {}",
            [gen_t2s(pattern), gen_t2s(term)],
            notify=True
        )

        return unify(term2string(term), pattern[1])

    elif pattern[0] == 'if-exp':

        # If we are evaluating subsumtion
        if not unifying:

            # If we are evaluating subsumption between two different conditional patterns
            # we want to 'punt' and print a warning message.
            if term[0] == 'if-exp':

                if not state.cond_warning:
                    warning("Redundant pattern detection is not supported for conditional pattern expressions.")
                    state.cond_warning = True

            # Otherwise if the term is not another cmatch the clauses are correctly ordered.
            raise PatternMatchFailed(
                "Subsumption relatioship broken, pattern will not be rendered redundant.")

        (IF_EXP, cond_exp, pexp, else_exp) = pattern

        if else_exp[0] != 'null':
            raise PatternMatchFailed("conditional patterns do not support 'else' clauses")

        # Explicit messaging
        message_explicit("Conditional match: if ({})",
            [gen_t2s(cond_exp)],
            notify=True, increase=True
        )

        unifiers = unify(term, pexp, unifying)

        if state.constraint_lvl:
            state.symbol_table.push_scope({})

        # evaluate the conditional expression in the
        # context of the unifiers.
        declare_unifiers(unifiers)
        bool_val = map2boolean(walk(cond_exp))

        if state.constraint_lvl:
            state.symbol_table.pop_scope()

        if bool_val[1]:
            message_explicit("Condition met, {}",
                [gen_t2s(cond_exp)], decrease=True
            )
            return unifiers
        else:
            message_explicit("Condition ({}) failed",
                [gen_t2s(cond_exp)], decrease=True
            )
            raise PatternMatchFailed(
                "conditional pattern match failed")

    elif term[0] == 'if-exp':
        # We will only get here when evaluating subsumption

        # If we get here, a conditional pattern clause is placed after a non-conditonal
        # pattern clause. Therefore, we need to check if the subsume because if they do
        # the conditonal clause is redundant.
        (IF_EXP, cond_exp, pexp, else_exp) = term

        if else_exp[0] != 'null':
            raise PatternMatchFailed("conditional patterns do not support 'else' clauses")

        #return unify(pexp,pattern,False)
        # Otherwise if the term is not another cmatch the clauses are correctly ordered.
        raise PatternMatchFailed(
            "conditional patterns not supported.")

    elif pattern[0] == 'typematch':
        typematch = pattern[1]
        nextIndex = 0 #indicates index of where we will 'look' next
        
        message_explicit("Typematch {} to type {}", [gen_t2s(term), typematch])

        if typematch in ['string','real','integer','list','tuple','boolean','none']:

            if (not unifying):

                #walk a different path for this node
                if (term[0] == 'typematch'):
                    nextIndex = 1

                #handle lists/head-tails subsuming each other
                if (term[0] in ["list","head-tail"]):
                    if ((typematch == 'list')):
                        return []

            if typematch == term[nextIndex]:
                message_explicit("Success!", level="secondary")
                return []
            else:
                message_explicit("Failure", level="secondary")
                raise PatternMatchFailed(
                    "expected type '{}' got a term of type '{}'"
                    .format(typematch, term[nextIndex]))

        elif typematch == 'function':
            # matching function and member function values
            if term[0] in ['function-val','member-function-val']:
                message_explicit("Success!", level="secondary")
                return []
            else:
                message_explicit("Failure", level="secondary")
                raise PatternMatchFailed(
                    "expected type '{}' got a term of type '{}'"
                    .format(typematch, term[0]))

        elif typematch == 'pattern':
            if unifying:

                # any kind of structure can be a pattern, and variables
                # see globals.py for a definition of 'patterns'
                if term[nextIndex] in patterns:
                    message_explicit("Success!", level="secondary")
                    return []
                else:
                    message_explicit("Failure", level="secondary")
                    raise PatternMatchFailed(
                            "expected type '{}' got a term of type '{}'"
                            .format(typematch, term[0]))

            else: # Evaluating typematch-pattern subsumption
                #walk a different path for this one node
                if (term[0] == 'typematch'):
                    nextIndex = 1

                #handle lists/head-tails subsuming each other
                if (term[0] in ["list","head-tail"]):
                    if ((typematch == 'list')):
                        message_explicit("Success!", level="secondary")
                        return []

                if term[nextIndex] in pattern_subsumes:
                    message_explicit("Success!", level="secondary")
                    return []
                else:
                    message_explicit("Failure", level="secondary")
                    raise PatternMatchFailed(
                        "expected type '{}' got a term of type '{}'"
                        .format(typematch, term[nextIndex]))

        elif term[0] == 'object':
            if state.symbol_table.lookup_sym(typematch)[0] != 'struct':
                raise PatternMatchFailed( "'{}' is not a type".format(typematch) )

            (OBJECT,
                (STRUCT_ID, (ID, struct_id)),
                (OBJECT_MEMORY, LIST)) = term
            if struct_id == typematch:
                message_explicit("Success!", level="secondary")
                return []
            else:
                message_explicit("Failure", level="secondary", increase=True)
                raise PatternMatchFailed(
                    "expected type '{}' got an object of type '{}'"
                    .format(typematch, struct_id))

        else:
            if state.symbol_table.lookup_sym(typematch)[0] != 'struct':
                message_explicit("Failure", level="secondary", increase=True)

                raise PatternMatchFailed( "'{}' is not a type".format(typematch) )
            else:
                message_explicit("Failure", level="secondary", increase=True)

                raise PatternMatchFailed(
                    "expected type '{}' got an object of type '{}'"
                    .format(typematch, term[0]))


    elif pattern[0] == 'named-pattern':
        # unpack pattern
        (NAMED_PATTERN, name_exp, p) = pattern
        message_explicit("Matching term {} and pattern {} to [{}]",
            [gen_t2s(term), gen_t2s(p), gen_t2s(name_exp)],
            notify=True, increase=True
        )
        # name_exp can be an id or an index expression.
        unifiers = unify(term, p, unifying) + [(name_exp, term)]
        
        message_explicit("Matched ({} and {})", 
            [gen_t2s(unifiers[0][0]), gen_t2s(unifiers[0][1])],
            decrease=True
        )

        return unifiers

    elif pattern[0] == 'none':
        if term[0] != 'none':
            message_explicit("{} and none do not match",
                [gen_t2s(term)]
            )
            raise PatternMatchFailed("expected 'none' got '{}'"
                    .format(term[0]))
        else:
            message_explicit("None and None match", level="secondary")
            return []

    # NOTE: functions/foreign are allowed in terms as long as they are matched
    # by a variable in the pattern - anything else will fail
    elif term[0] in (unify_not_allowed - {'function-val', 'foreign'}):
        raise PatternMatchFailed(
            "term of type '{}' not allowed in pattern matching"
            .format(term[0]))

    elif pattern[0] in unify_not_allowed:
        raise PatternMatchFailed(
            "pattern of type '{}' not allowed in pattern matching"
            .format(pattern[0]))

    elif pattern[0] == 'pattern':
        # pattern operator on the pattern side can always be ignored
        # --TODO double check - ttc
        if term[0] == 'pattern':
            return unify(term[1], pattern[1], unifying)
        else:
            return unify(term, pattern[1], unifying)

    elif term[0] == 'pattern' and pattern[0] not in ['id', 'index']:
        # ignore pattern operator on the term if we are not trying to unify term with
        # a variable or other kind of lval
        return unify(term[1], pattern, unifying)

    elif term[0] == 'object' and pattern[0] == 'apply':
        # unpack term
        (OBJECT,
         (STRUCT_ID, (ID, struct_id)),
         (OBJECT_MEMORY, (LIST, obj_memory))) = term
        # unpack pattern
        (APPLY,
         (ID, apply_id),
         arg) = pattern
        type = state.symbol_table.lookup_sym(apply_id)
        if type[0] != 'struct':
            raise PatternMatchFailed("'{}' is not a type".format(apply_id))

        if struct_id != apply_id:
            raise PatternMatchFailed("expected type '{}' got type '{}'"
                .format(apply_id, struct_id))
        # we are comparing raw lists here
        if arg[0] == 'tuple':
            pattern_list = arg[1]
        else:
            pattern_list = [arg]

        # only pattern match on object data members
        message_explicit("Matching object {}{} and pattern {}{}",
            [apply_id,  gen_t2s( ('tuple', data_only(obj_memory)) ),
            struct_id, gen_t2s( ('tuple', pattern_list) )],
            notify=True
        )
        # Running through the list elements indivuidually allows for
        # better debugging information
        unifiers = []
        data = data_only(obj_memory)

        for i in range(len(data)):
            unifiers += unify(data[i], pattern_list[i], unifying)

        return unifiers

    elif pattern[0] == 'index': # list element lval access
        unifier = (pattern, term)
        return [unifier]

# lhh: looking at patterns as values we are now allowed to match
# against patterns as terms where the terms now include variables.
#    elif term[0] == 'id' and unifying: # variable in term not allowed
#        raise PatternMatchFailed(      # when unifying
#            "variable '{}' in term not allowed"
#            .format(term[1]))

    elif pattern[0] == 'id': # variable in pattern add to unifier
        sym = pattern[1]
        if sym == '_': # anonymous variable - ignore unifier
            return []
        else:
            unifier = (pattern, term)
            return [unifier]

    elif pattern[0] in ['head-tail', 'raw-head-tail']:
        message_explicit("Matching {} to {}",
            [gen_t2s(term), gen_t2s(pattern)],
            notify=True
        )
        # if we are unifying or we are not evaluating subsumption
        #  to another head-tail
        if unifying or term[0] not in ['head-tail', 'raw-head-tail']:
            (HEAD_TAIL, pattern_head, pattern_tail) = pattern
            (LIST, list_val) = term

            if LIST != 'list':
                message_explicit("Failed", level="secondary")
                raise PatternMatchFailed(
                    "head-tail operator expected type 'list' got type '{}'"
                    .format(LIST))

            if not len(list_val):
                message_explicit("Failed", level="secondary")

                raise PatternMatchFailed(
                    "head-tail operator expected a non-empty list")

            list_head = list_val[0]
            list_tail = ('list', list_val[1:])
            unifier = []

            unifier += unify(list_head, pattern_head, unifying)
            unifier += unify(list_tail, pattern_tail, unifying)

            check_repeated_symbols(unifier) #Ensure we have no non-linear patterns
            message_explicit("Success!", level="secondary")

            return unifier

        else: #Else we are evaluating subsumption to another head-tail

            lengthH = head_tail_length(pattern) #H->higher order of predcence pattern
            lengthL = head_tail_length(term)    #L->lower order of predcence pattern

            if lengthH == 2 and lengthL != 2:
                return unify(pattern[1],term[1],unifying)

            if (lengthH > lengthL): # If the length of the higher presedence pattern is greater
                                    # then length of the lower precedence pattern, it is not redundant
                message_explicit("Failed", level="secondary")
                raise PatternMatchFailed(
                    "Subsumption relatioship broken, pattern will not be rendered redundant.")

            else: #Else we continue evaluating the different terms in the head-tail pattern
                (HEAD_TAIL, patternH_head, patternH_tail) = pattern
                (HEAD_TAIL, patternL_head, patternL_tail) = term

                unifier = []
                for i in range(lengthH):
                    unifier += unify(patternL_head,patternH_head,unifying)
                    try:
                        (RAW_HEAD_TAIL, patternH_head, patternH_tail) = patternH_tail
                        (RAW_HEAD_TAIL, patternL_head, patternL_tail) = patternL_tail
                    except:
                        break

                check_repeated_symbols(unifier) #Ensure we have no non-linear patterns
                message_explicit("Success!", level="secondary")
                return unifier

    elif pattern[0] == 'deref':  # ('deref', v)
        # v can be an AST representing any computation
        # that produces a pattern.

        message_explicit("Dereferencing {}", [gen_t2s(pattern[1])])

        p = walk(pattern[1])

        message_explicit("{} -> {}", [gen_t2s(pattern), gen_t2s(p)], 
            level="secondary")

        notify_explicit()

        return unify(term,p,unifying)

    # builtin operators look like apply lists with operator names
    elif pattern[0] == 'apply':
        if term[0] != pattern[0]: # make sure both are applys
            raise PatternMatchFailed(
                "term and pattern disagree on structure")

        # unpack the apply structures
        (APPLY, (ID, t_id), t_arg) = term
        (APPLY, (ID, p_id), p_arg) = pattern

        # make sure apply id's match
        if t_id != p_id:
            raise PatternMatchFailed(
                "term '{}' does not match pattern '{}'"
                .format(t_id, p_id))

        # unify the args
        return unify(t_arg, p_arg, unifying)

    elif pattern[0] == 'constraint':
        message_explicit("[Begin] constraint pattern: {}",
            [gen_t2s(pattern[1])],
            notify=True, increase=True
        )

        state.constraint_lvl += 1
        unifier = unify(term,pattern[1])
        state.constraint_lvl -= 1

        message_explicit("[End] constraint pattern", decrease=True)
        return [] #Return an empty unifier

    elif term[0] != pattern[0]:  # nodes are not the same
        message_explicit("Fail: {} and {} are not the same",
            [term[0], pattern[0]], level="secondary")
        
        raise PatternMatchFailed(
            "nodes '{}' and '{}' are not the same"
            .format(term[0], pattern[0]))

    elif len(term) != len(pattern): # nodes are not of same the arity
        message_explicit("Fail: {} and {} are not the same arity",
            [term[0], pattern[0]], level="secondary")

        raise PatternMatchFailed(
            "nodes '{}' and '{}'' are not of the same arity"
            .format(term[0], pattern[0]))

    else:
        #lhh
        #print("unifying {}".format(pattern[0]))
        unifier = []
        for i in range(1,len(term)):
            unifier += unify(term[i], pattern[i], unifying)
        #lhh
        #print("returning unifier: {}".format(unifier))
        return unifier


#########################################################################
def eval_actual_args(args):

    return walk(args)

#########################################################################
def declare_formal_args(unifiers):
    # unfiers is of the format: [ (pattern, term), (pattern, term),...]

    for u in unifiers:
        (pattern, term) = u
        (ID, sym) = pattern # in unifiers the pattern is always a variable
        assert_match(ID,'id')
        if sym == 'this':
            raise ValueError("'this' is a reserved word")
        state.symbol_table.enter_sym(sym, term)

#########################################################################
# Evaluates a set of unifiers for the presence of repeated variable
# names within a pattern. Repeated variables names within the same pattern
# are what is called a non-linear pattern, which is not currently supported
# by Asteroid.
# This function will raise a NonLinearPatternError exception when a non-linear
# pattern has been recognized.
# Otherwise, this function returns control to the caller after finishing.
def check_repeated_symbols( unifiers ):

    symbols = {} # Will hold all previously seen unifiers(term-pattern) as (key-value)
    skip_unifier = False #Determines if we want to eval the current unifier pair

    # For each pair of unifiers
    for unifier in unifiers:

        # Unpack the pattern-term pair
        (pattern, term) = unifier

        # If the pattern side is an ID node, unpack it
        # Else we skip this unifier
        if pattern[0] == 'id':
            (ID, sym) = pattern
        else:
            skip_unifier = True

        # If we are not skipping this turn, check to see if we have seen this
        # variable before.
        if skip_unifier:
            skip_unifier = False

        elif sym in symbols: # We have found a non-linear pattern
            raise NonLinearPatternError(
            "multiple instances of '{}' found within pattern.".format(sym))

        else: # Else we have never seen this before so we record it.
            symbols[sym] = term

#########################################################################
# we are indexing into the memory of either a list/tuple/string or an
# object to read the memory.
#
# NOTE: when indexed with a scalar it will return a single value,
# that value of course could be a list etc.  When index with a list
# then it will return a list of values. Therefore:
#       a@1 =/= a@[1]
# the value on the left of the inequality is a single value, the
# value on the right is a singleton list.
def read_at_ix(structure_val, ix):

    # find the actual memory we need to access
    # list: return the actual list
    if structure_val[0] in ['list', 'tuple', 'string']:
        if structure_val[0] == 'list' \
        and ix[0] == 'id' \
        and ix[1] in list_member_functions:
            # we are looking at the function name of a list member
            # function - find the implementation and return it.
            impl_name = list_member_functions[ix[1]]
            # remember the object reference.
            return ('member-function-val',
                    structure_val,
                    state.symbol_table.lookup_sym(impl_name))
        elif structure_val[0] == 'string' \
        and ix[0] == 'id' \
        and ix[1] in string_member_functions:
            # we are looking at the function name of a string member
            # function - find the implementation and return it.
            impl_name = string_member_functions[ix[1]]
            # remember the object reference.
            return ('member-function-val',
                    structure_val,
                    state.symbol_table.lookup_sym(impl_name))
        else:
            # get a reference to the memory
            memory = structure_val[1]
            # compute the index
            ix_val = walk(ix)

    # for objects we access the object memory
    elif structure_val[0] == 'object':
        (OBJECT,
         (STRUCT_ID, (ID, struct_id)),
         (OBJECT_MEMORY, (LIST, memory))) = structure_val
        # compute the index -- for objects this has to be done
        # in the context of the struct scope
        struct_val = state.symbol_table.lookup_sym(struct_id)
        # unpack the struct value
        (STRUCT,
         (MEMBER_NAMES, (LIST, member_names)),
         (STRUCT_MEMORY, (LIST, struct_memory))) = struct_val

        if ix[0] == 'id' and ix[1] in member_names:
            ix_val = ('integer', member_names.index(ix[1]))
        else:
            ix_val = walk(ix)

    elif structure_val[0] == 'pattern':
        # simple patterns are just structures - skip the pattern operator
        return read_at_ix(structure_val[1], ix)

    else:
        raise ValueError("term '{}' is not indexable"
                         .format(term2string(structure_val)))

    # index into memory and get value(s)
    if ix_val[0] == 'integer':
        if structure_val[0] == 'string':
            return ('string', memory[ix_val[1]])
        elif structure_val[0] == 'object' \
        and memory[ix_val[1]][0] == 'function-val':
            # remember the object reference.
            return ('member-function-val',
                    structure_val,
                    memory[ix_val[1]])
        else:
            return memory[ix_val[1]]

    elif ix_val[0] == 'list':
        if len(ix_val[1]) == 0:
            raise ValueError("index list is empty")

        return_memory = []
        for i in ix_val[1]:
            (IX_EXP_TYPE, ix_exp) = i

            if IX_EXP_TYPE == 'integer':
                return_memory.append(memory[ix_exp])
            else:
                raise ValueError("unsupported list index")

        if structure_val[0] == 'string':
            return ('string',"".join(return_memory))
        else:
            return ('list', return_memory)

    else:
        raise ValueError("index op '{}' not supported".format(ix_val[0]))

#########################################################################
# we are indexing into the memory of either a list or an object to
# write into the memory.
def store_at_ix(structure_val, ix, value):
    # find the actual memory we need to access
    # for lists it is just the python list
    if structure_val[0] == 'list':
        memory = structure_val[1]
        # compute the index
        ix_val = walk(ix)

    # for objects we access the object memory
    elif structure_val[0] == 'object':
        (OBJECT,
         (STRUCT_ID, (ID, struct_id)),
         (OBJECT_MEMORY, (LIST, memory))) = structure_val
        # compute the index -- for objects this has to be done
        # in the context of the struct scope
        struct_val = state.symbol_table.lookup_sym(struct_id)
        # unpack the struct value
        (STRUCT,
         (MEMBER_NAMES, (LIST, member_names)),
         (STRUCT_MEMORY, (LIST, struct_memory))) = struct_val

        if ix[0] == 'id' and ix[1] in member_names:
            ix_val = ('integer', member_names.index(ix[1]))
        else:
            ix_val = walk(ix)

    elif structure_val[0] == 'pattern':
        # simple patterns are just structures - skip the pattern operator
        return store_at_ix(structure_val[1], ix, value)

    else:
        raise ValueError("term '{}' is not a mutable structure"
                         .format(term2string(structure_val)))

    # Next, we do the actual memory storage operation

    # If it's just an integer, index into that location and
    # set the value
    if ix_val[0] == 'integer':
        memory[ix_val[1]] = value
        return

    # otherwhise, if the index is a list
    elif ix_val[0] == 'list':

        # Make sure the rval is a list
        if value[0] != 'list':
            raise ValueError('pattern slicing needs values to be a list')
        elif value[0] == 'list' and (len(ix_val[1]) != len(value[1])):
            raise ValueError('pattern slicing needs indexes and values of equal length')

        # Get the l/rval
        (LIST, lval) = ix_val
        (LIST_r, rval) = value

        # For each index
        for i in range(len(lval)):
            # Get the memory location of the lval and set it to the
            # corresponding rval value
            (INTEGER, location) = lval[i]
            memory[location] = rval[i]

        return

    else:
        raise ValueError("index op '{}' in patterns not supported"
                         .format(ix_val[0]))

#########################################################################
def handle_builtins(node):
    (APPLY, (ID, opname), args) = node
    assert_match(APPLY, 'apply')
    assert_match(ID, 'id')

    if opname in binary_operators:
        (TUPLE, bin_args)= args
        val_a = walk(bin_args[0])
        val_b = walk(bin_args[1])

        if opname == '__plus__':
            type = promote(val_a[0], val_b[0])
            if type in ['integer', 'real', 'list', 'boolean']:
                return (type, val_a[1] + val_b[1])
            elif type == 'string':
                return (type, term2string(val_a) + term2string(val_b))
            else:
                raise ValueError("unsupported type '{}' in +".format(type))
        elif opname == '__minus__':
            type = promote(val_a[0], val_b[0])
            if type in ['integer', 'real']:
                return (type, val_a[1] - val_b[1])
            else:
                raise ValueError("unsupported type '{}' in -".format(type))
        elif opname == '__times__':
            type = promote(val_a[0], val_b[0])
            if type in ['integer', 'real']:
                return (type, val_a[1] * val_b[1])
            else:
                raise ValueError('unsupported type in *')
        elif opname == '__divide__':
            type = promote(val_a[0], val_b[0])
            if type == 'integer':
                return (type, int(val_a[1]) // int(val_b[1]))
            elif type == 'real':
                return ('real', float(val_a[1]) / float(val_b[1]))
            else:
                raise ValueError('unsupported type in /')
        elif opname == '__or__':
            # NOTE: do we need to typecheck here?
            if map2boolean(val_a)[1] == True or map2boolean(val_b)[1] == True:
               return ('boolean', True)
            else:
               return ('boolean', False)
        elif opname == '__and__':
            # NOTE: do we need to typecheck here?
            if map2boolean(val_a)[1] == True and map2boolean(val_b)[1] == True:
               return ('boolean', True)
            else:
               return ('boolean', False)
        elif opname == '__eq__':
            type = promote(val_a[0], val_b[0])
            if type in ['integer', 'real', 'list', 'tuple', 'boolean', 'none']:
                return ('boolean', val_a[1] == val_b[1])
            elif type == 'string':
                return ('boolean', term2string(val_a) == term2string(val_b))
            else:
                raise ValueError('unsupported type in ==')
        elif opname  == '__ne__':
            type = promote(val_a[0], val_b[0])
            if type in ['integer', 'real', 'list', 'tuple', 'boolean', 'none']:
                return ('boolean', val_a[1] != val_b[1])
            elif type == 'string':
                return ('boolean', term2string(val_a) != term2string(val_b))
            else:
                raise ValueError('unsupported type in =/=')
        elif opname == '__le__':
            type = promote(val_a[0], val_b[0])
            if type in ['integer', 'real']:
                return ('boolean', val_a[1] <= val_b[1])
            else:
                raise ValueError('unsupported type in <=')
        elif opname == '__lt__':
            type = promote(val_a[0], val_b[0])
            if type in ['integer', 'real']:
                return ('boolean', val_a[1] < val_b[1])
            else:
                raise ValueError('unsupported type in <')
        elif opname == '__ge__':
            type = promote(val_a[0], val_b[0])
            if type in ['integer', 'real']:
                return ('boolean', val_a[1] >= val_b[1])
            else:
                raise ValueError('unsupported type in >=')
        elif opname == '__gt__':
            type = promote(val_a[0], val_b[0])
            if type in ['integer', 'real']:
                return ('boolean', val_a[1] > val_b[1])
            else:
                raise ValueError('unsupported type in >')
        else:
            raise ValueError("unknown builtin binary operation '{}'".format(opname))

    elif opname in unary_operators:
        arg_val = walk(args)

        if opname == '__not__':
            val = map2boolean(arg_val)
            if val[1] == False:
                return ('boolean', True)
            elif val[1] == True:
                return ('boolean', False)
            else:
                raise ValueError("not a boolean value in 'not'")
        elif opname == '__uminus__':
            if arg_val[0] in ['integer', 'real']:
                return (arg_val[0], - arg_val[1])
            else:
                raise ValueError(
                    "unsupported type '{}' in unary minus"
                    .format(arg_val[0]))
        elif opname == '__uplus__':
            if arg_val[0] in ['integer', 'real']:
                return (arg_val[0], + arg_val[1])
            else:
                raise ValueError(
                    "unsupported type '{}' in unary plus"
                    .format(arg_val[0]))
        else:
            raise ValueError("unknown builtin unary operation '{}'".format(opname))

#########################################################################
def handle_call(obj_ref, fval, actual_val_args, fname):
    # Needed for later
    global debugging

    # function calls transfer control - save our caller's lineinfo
    # we save the debug information here to preserve lineinfo between
    # function calls between files.
    old_lineinfo = state.lineinfo

    (FUNCTION_VAL, body_list, closure) = fval
    assert_match(FUNCTION_VAL, 'function-val')

    state.trace_stack.append((state.lineinfo[0],
                              state.lineinfo[1],
                              fname))

    # static scoping for functions
    # Note: we have to do this here because unifying
    # over the body patterns can introduce variable declarations,
    # think conditional pattern matching.
    state.symbol_table.saved_configs.append(
        state.symbol_table.get_config()
    )

    state.symbol_table.set_config(closure)
    state.symbol_table.push_scope({})

    # Explicit message
    message_explicit("Call: {}({})",
        [fname, gen_t2s(actual_val_args)],
        increase=True
    )

    # We want to return back to whatever tab level the function
    # call started at, so, we grab this here to reset to at
    # the return. This allows recursive functions to have
    # the proper formatting
    if debugging: cur_tab_level = debugger.tab_level

    # iterate over the bodies to find one that unifies with the actual parameters
    (BODY_LIST, (LIST, body_list_val)) = body_list
    unified = False

    for i in range(0, len(body_list_val), 2):
        # Process lineinfo
        lineinfo = body_list_val[ i ]
        process_lineinfo(lineinfo)

        # Deconstruct function body
        (BODY,
        (PATTERN, p),
        (STMT_LIST, stmts)) = body_list_val[ i + 1]

        message_explicit("Attempting to match {} with pattern {}",
            [gen_t2s(actual_val_args), gen_t2s(p)], level="primary",
            notify=True, increase=True
        )
        
        try:
            # Attempt to unify the actual args and the pattern
            unifiers = unify(actual_val_args, p)
            unified = True

            # Do our explicit message
            message_explicit("Success! Matched function body", level="primary", decrease=True)

        except PatternMatchFailed:
            # Reset the tab level
            if explicit_enabled():
                debugger.tab_level = cur_tab_level + 1

            # Print the explicit messaging
            message_explicit("Failed to match function body", level="tertiary", decrease=True)

            unifiers = []
            unified = False

        if unified:
            break

    # Reset the tab level back to the function call's level
    if debugging: debugger.tab_level = cur_tab_level

    if not unified:
        raise ValueError("actual argument '{}' not recognized by function '{}'"
                         .format(term2string(actual_val_args),fname))
    declare_formal_args(unifiers)

    # if we have an obj reference bind it to the
    # variable 'this'
    if obj_ref:
        state.symbol_table.enter_sym('this', obj_ref)

    # OWM: The following segment is a repeat of the bottom of this function.
    # We need to do this because redundant patterns can break scope and
    # some debugger and state features as they exit computation.

    # Check for useless patterns
    try:
        if state.eval_redundancy:
            old_debugging = debugging
            debugging = False
            check_redundancy(body_list, fname)
            debugging = old_debugging

    # Reset settings
    except RedundantPatternFound as r:
        debugging = old_debugging

        # coming back from a function call - restore caller's env
        state.lineinfo = old_lineinfo

        # Keep debugger up to date
        if debugging:
            debugger.set_lineinfo(state.lineinfo)

        state.symbol_table.pop_scope()
        state.symbol_table.set_config(state.symbol_table.saved_configs.pop())

        state.trace_stack.pop()

        raise r

    # execute the function
    global function_return_value

    try:
        function_return_value.append(None)

        walk_stmt_list(stmts)

        # Pop the return value
        val = function_return_value.pop()

        # If we have one, set it as the retval
        if val:
            return_value = val

        # Otherwhise default the return value to none
        else:
            return_value = ('none', None)

    except ReturnValue as val:
        # we got here because a return statement threw a return object
        function_return_value.pop()
        return_value = val.value

    # coming back from a function call - restore caller's env
    state.lineinfo = old_lineinfo

    # Keep debugger up to date
    if debugging: debugger.set_lineinfo(state.lineinfo)

    state.symbol_table.pop_scope()
    state.symbol_table.set_config(state.symbol_table.saved_configs.pop())

    state.trace_stack.pop()

    message_explicit("Return: {} from {}",
            [("None" if (not return_value[1]) else gen_t2s(return_value)), 
            fname],
            decrease=True
    )

    return return_value

#########################################################################
def declare_unifiers(unifiers):
    # walk the unifiers and bind name-value pairs into the symtab

    # TODO: check for repeated names in the unfiers
    for unifier in unifiers:

        #lhh
        #print("unifier: {}".format(unifier))
        (lval, value) = unifier

        if lval[0] == 'id':
            if lval[1] == 'this':
                raise ValueError("'this' is a reserved word")
            state.symbol_table.enter_sym(lval[1], value)

        elif lval[0] == 'index': # list/structure lval access
            # Note: structures have to be declared before index access
            # can be successful!!  They have to be declared so that there
            # is memory associated with the structure.

            (INDEX, structure, ix) = lval
            # look at the semantics of 'structure'
            structure_val = walk(structure)
            # indexing/slicing
            # update the memory of the object.
            store_at_ix(structure_val, ix, value)

        else:
            raise ValueError("unknown unifier type '{}'".format(lval[0]))
    
    # Explicit messaging for unifiers
    if explicit_enabled():
        # if we've unified anything
        if unifiers:
            # Create our format string and list of terms
            fstring = ""
            terms = []

            # For each unifier except the last one
            for (lval, value) in unifiers[:-1]:
                # Add to our format string and our terms
                fstring += "{} = {}, "
                terms += [gen_t2s(lval), gen_t2s(value)]

            # Get the last unifier and add it to the terms and fstring
            (lval, value) = unifiers[-1]
            fstring += "{} = {}"
            terms += [gen_t2s(lval), gen_t2s(value)]

            # Print our message
            message_explicit(fstring, terms)

#########################################################################
# node functions
#########################################################################
def global_stmt(node):
    notify_debugger()

    (GLOBAL, (LIST, id_list)) = node
    assert_match(GLOBAL, 'global')
    assert_match(LIST, 'list')

    global_str = ""

    for id_tuple in id_list:
        (ID, id_val) = id_tuple
        if state.symbol_table.is_symbol_local(id_val):
            raise ValueError("'{}' is already local, cannot be declared global"
                             .format(id_val))
        state.symbol_table.enter_global(id_val)
        global_str += "{}, ".format(id_val)
    
    message_explicit("Global defs: {}", [global_str[:-1]] )

#########################################################################
def assert_stmt(node):
    notify_debugger()

    (ASSERT, exp) = node
    assert_match(ASSERT, 'assert')

    exp_val = walk(exp)

    # mapping asteroid assert into python assert
    assert exp_val[1], 'assert failed'
#########################################################################
def unify_stmt(node):
    notify_debugger()

    (UNIFY, pattern, exp) = node
    assert_match(UNIFY, 'unify')

    term = walk(exp)

    unifiers = unify(term, pattern)

    declare_unifiers(unifiers)

#########################################################################
def return_stmt(node):
    notify_debugger()

    (RETURN, e) = node
    assert_match(RETURN, 'return')
    
    retval = walk(e)

    raise ReturnValue(retval)

#########################################################################
def break_stmt(node):
    notify_debugger()

    (BREAK,) = node
    assert_match(BREAK, 'break')

    raise Break()

#########################################################################
def throw_stmt(node):
    notify_debugger()

    (THROW, object) = node
    assert_match(THROW, 'throw')
    
    throw_object = walk(object)

    raise ThrowValue(throw_object)

#########################################################################
def try_stmt(node):
    notify_debugger()

    (TRY,
     (STMT_LIST, try_stmts),
     (CATCH_LIST, (LIST, catch_list))) = node

    stepping = debugger_has_stepped()
    try:
        walk_stmt_list(try_stmts)

    # NOTE: in Python the 'as inst' variable is only local to the catch block???
    # NOTE: we map user visible Python exceptions into standard Asteroid exceptions
    #       by constructing Exception objects - see prologue.ast
    except ThrowValue as inst:
        except_val = inst.value
        inst_val = inst

    except ReturnValue as inst:
        # return values should never be captured by user level try stmts - rethrow
        raise inst

    except PatternMatchFailed as inst:
        except_val = ('object',
                         ('struct-id', ('id', 'Exception')),
                         ('object-memory',
                          ('list',
                           [('string', 'PatternMatchFailed'),
                            ('string', inst.value)])))
        inst_val = inst

    except RedundantPatternFound as inst:
        except_val = ('object',
                         ('struct-id', ('id', 'Exception')),
                         ('object-memory',
                          ('list',
                           [('string', 'RedundantPatternFound'),
                            ('string', str(inst))])))
        inst_val = inst

    except NonLinearPatternError as inst:
        except_val = ('object',
                         ('struct-id', ('id', 'Exception')),
                         ('object-memory',
                          ('list',
                           [('string', 'NonLinearPatternError'),
                            ('string', str(inst))])))
        inst_val = inst

    except ArithmeticError as inst:
        except_val = ('object',
                         ('struct-id', ('id', 'Exception')),
                         ('object-memory',
                          ('list',
                           [('string', 'ArithmeticError'),
                            ('string', str(inst))])))
        inst_val = inst

    except FileNotFoundError as inst:
        except_val = ('object',
                         ('struct-id', ('id', 'Exception')),
                         ('object-memory',
                          ('list',
                           [('string', 'FileNotFound'),
                            ('string', str(inst))])))
        inst_val = inst

    except Exception as inst:
        # mapping general Python exceptions into Asteroid's SystemError
        except_val = ('object',
                         ('struct-id', ('id', 'Exception')),
                         ('object-memory',
                          ('list',
                           [('string', 'SystemError'),
                            ('string', str(inst))])))
        inst_val = inst

    else:
        # no exceptions found in the try statements
        return

    # we had an exception - walk the catch list and find an appropriate set of
    # catch statements.
    for catch_val in catch_list:
        (CATCH,
         (CATCH_PATTERN, catch_pattern),
         (CATCH_STMTS, catch_stmts)) = catch_val
        try:
            unifiers = unify(except_val, catch_pattern)
        except PatternMatchFailed:
            pass
        else:
            declare_unifiers(unifiers)
            walk_stmt_list(catch_stmts, step_state=stepping)
            return

    # no exception handler found - rethrow the exception
    raise inst_val

#########################################################################
def loop_stmt(node):
    notify_debugger()

    (LOOP, body_stmts) = node
    assert_match(LOOP, 'loop')

    (STMT_LIST, body) = body_stmts

    try:
        stepping = debugger_has_stepped()
        while True:
            walk_stmt_list(body, step_state=stepping)
    except Break:
        pass

#########################################################################
def while_stmt(node):
    notify_debugger()

    (WHILE, cond_exp, body_stmts) = node
    assert_match(WHILE, 'while')

    (COND_EXP, cond) = cond_exp
    (STMT_LIST, body) = body_stmts

    try:
        stepping = debugger_has_stepped()

        (COND_TYPE, cond_val) = map2boolean(walk(cond))
        while cond_val:
            walk_stmt_list(body, step_state=stepping)
            (COND_TYPE, cond_val) = map2boolean(walk(cond))
    except Break:
        pass

#########################################################################
def repeat_stmt(node):
    notify_debugger()

    (REPEAT, body_stmts, cond_exp) = node
    assert_match(REPEAT, 'repeat')

    (COND_EXP, cond) = cond_exp
    (STMT_LIST, body) = body_stmts

    try:
        stepping = debugger_has_stepped()
        while True:
            walk_stmt_list(body, step_state=stepping)
            (COND_TYPE, cond_val) = map2boolean(walk(cond))
            if cond_val:
                break

    except Break:
        pass

#########################################################################
def for_stmt(node):
    notify_debugger()

    (FOR, (IN_EXP, in_exp), (STMT_LIST, stmt_list)) = node
    assert_match(FOR, 'for')

    (IN, pattern, list_term) = in_exp

    # expand the list_term
    (LIST_TYPE, list_val) = walk(list_term)
    if LIST_TYPE not in ['list','string']:
        raise ValueError("only iteration over strings and lists is supported")

    # we allow iteration over two types of structures: (1) lists (2) strings
    # if it is a string turn the list_val into a list of Asteroid characters.
    if LIST_TYPE == 'string':
        new_list = []
        for c in list_val:
            new_list.append(('string',c))
        list_val = new_list

    # for each term on the list unfiy with pattern, declare the bound variables,
    # and execute the loop body in that context
    # NOTE: just like Python, loop bodies do not create a new scope!
    # NOTE: we can use unification as a filter of elements:
    #
    #      for (2,y) in [(1,11), (1,12), (1,13), (2,21), (2,22), (2,23)]  do
    #             print y.
    #      end for
    try:
        stepping = debugger_has_stepped()
        for term in list_val:
            try:
                unifiers = unify(term,pattern)
            except PatternMatchFailed:
                pass
            else:
                declare_unifiers(unifiers)
                walk_stmt_list(stmt_list, step_state=stepping)
                
    except Break:
        pass

#########################################################################
def if_stmt(node):
    notify_debugger()

    (IF, (LIST, if_list)) = node
    assert_match(IF, 'if')
    assert_match(LIST, 'list')

    for i in range(0,len(if_list),2):

        lineinfo = if_list[ i ]
        process_lineinfo(lineinfo)

        (IF_CLAUSE,
         (COND, cond),
         (STMT_LIST, stmts)) = if_list[ i + 1 ]

        (BOOLEAN, cond_val) = map2boolean(walk(cond))

        if cond_val:
            walk_stmt_list(stmts)
            break
#########################################################################
def struct_def_stmt(node):
    notify_debugger()

    (STRUCT_DEF, (ID, struct_id), (MEMBER_LIST, (LIST, member_list))) = node
    assert_match(STRUCT_DEF, 'struct-def')
    assert_match(ID, 'id')
    assert_match(MEMBER_LIST, 'member-list')
    assert_match(LIST, 'list')

    # declare members
    # member names are declared as variables whose value is the slot
    # in a struct object
    struct_memory = [] # this will serve as a template for instanciating objects
    member_names = []

    for member_ix in range(len(member_list)):
        member = member_list[member_ix]
        if member[0] == 'data':
            (DATA, (ID, member_id)) = member
            struct_memory.append(('none', None))
            member_names.append(member_id)

        elif member[0] == 'unify':
            (UNIFY, (ID, member_id), function_exp) = member
            # Note: we have to bind a function VALUE into the structure memory
            function_val = walk(function_exp)
            struct_memory.append(function_val)
            member_names.append(member_id)

        elif member[0] == 'noop':
            pass
        else:
            raise ValueError("unsupported struct member '{}'".format(member[0]))

    struct_type = ('struct',
                  ('member-names', ('list', member_names)),
                  ('struct-memory', ('list', struct_memory)))

    state.symbol_table.enter_sym(struct_id, struct_type)

#########################################################################
def import_stmt(node):
    notify_debugger()

    global debugging
    old_debugging = debugging
    debugging = False

    (LIST, inlist) = node
    assert_match(LIST, 'import_stmt')

    for e in inlist:
        walk(e)

    debugging = old_debugging

    return

#########################################################################
def top_level_exp_stmt(node):

    (TOP_LEVEL_EXP, exp) = node
    assert_match(TOP_LEVEL_EXP, 'top-level-exp')

    notify_debugger()

    return walk(exp)

#########################################################################
def eval_exp(node):

    (EVAL, exp) = node
    assert_match(EVAL, 'eval')
    # lhh
    #print("in eval with {}".format(node))

    # Note: eval is essentially a macro call - that is a function
    # call without pushing a symbol table record.  That means
    # we have to first evaluate the argument to 'eval' before
    # walking the term.  This is safe because if the arg is already
    # the actual term it will be quoted and nothing happens if it is
    # a variable it will be expanded to the actual term.
    #lhh
    #print("before expand: {}".format(exp))
    exp_val_expand = walk(exp)
    #lhh
    #print("after expand: {}".format(exp_val_expand))
    # now walk the actual term
    state.ignore_pattern += 1
    exp_val = walk(exp_val_expand)
    #lhh
    #print("after walk: {}".format(exp_val))
    state.ignore_pattern -= 1
    return exp_val

#########################################################################
def apply_exp(node):
    (APPLY, f, arg) = node
    assert_match(APPLY, 'apply')

    # handle builtin operators that look like apply lists.
    if f[0] == 'id' and f[1] in operator_symbols:
        return handle_builtins(node)

    # handle function application
    # retrieve the function name from the AST
    if f[0] in ['function-exp','apply']:
        # cannot use the function expression as a name,
        # could be a very complex computation. the apply
        # node means that the lambda function has to still be
        # computed.
        f_name = 'lambda'
    elif f[0] == 'index':
        # object member function
        (INDEX, ix, (ID, f_name)) = f
        # 'str' is necessary in case we use an index value
        # instead of a function name -- see regression test test085.ast
        f_name = "member function " + str(f_name)
    else:
        # just a regular function call
        (ID, f_name) = f

    # evaluate the function expression and the arguments
    f_val = walk(f)
    arg_val = walk(arg)

    # object member function
    # NOTE: object member functions are passed an object reference.
    if f_val[0] == 'member-function-val':
        (MEMBER_FUNCTION_VAL, obj_ref, function_val) = f_val
        # Note: lists and strings are objects/mutable data structures, they
        # have member functions defined in the Asteroid prologue.
        result = handle_call(obj_ref,
                             function_val,
                             arg_val,
                             f_name)

    # regular function call
    elif f_val[0] == 'function-val':
        result = handle_call(None, f_val, arg_val, f_name)

    # object constructor call
    elif f_val[0] == 'struct':
        (ID, struct_id) = f
        (STRUCT,
         (MEMBER_NAMES, (LIST, member_names)),
         (STRUCT_MEMORY, (LIST, struct_memory))) = f_val

        # create our object memory - memory cells now have initial values
        # we use structure memory as an init template
        object_memory = struct_memory.copy()
        # create our object
        obj_ref = ('object',
                   ('struct-id', ('id', struct_id)),
                   ('object-memory', ('list', object_memory)))
        # if the struct has an __init__ function call it on the object
        # NOTE: constructor functions do not have return values.

        if '__init__' in member_names:
            slot_ix = member_names.index('__init__')
            init_fval = struct_memory[slot_ix]
            # calling a member function - push struct scope
            handle_call(obj_ref,
                        init_fval,
                        arg_val,
                        f_name)
        # the struct does not have an __init__ function but
        # we have a constructor call with args, e.g. Foo(1,2)
        # try to apply a default constructor by copying the
        # values from the arg list to the data slots of the object
        elif arg_val[0] != 'none':
            if arg_val[0] != 'tuple':
                arg_array = [arg_val]
            else:
                arg_array = arg_val[1]
            data_memory = data_only(object_memory)
            if len(data_memory) != len(arg_array):
                raise ValueError(
                    "default constructor expected {} argument{} got {}"
                    .format(len(data_memory),
                            "" if len(data_memory) else "s",
                            len(arg_array)))
            # copy initializers into object memory
            data_ix = data_ix_list(object_memory)
            for (i,k) in zip(data_ix, range(0,len(data_memory))):
                object_memory[i] = arg_array[k]

        # return the new object
        result = obj_ref

    else:
        raise ValueError("term '{}' is not a function, did you forget the end-of-line period?"
                         .format(term2string(f_val)))

    return result

#########################################################################
def index_exp(node):

    (INDEX, structure, ix) = node
    assert_match(INDEX, 'index')

    # look at the semantics of 'structure'
    structure_val = walk(structure)

    # indexing/slicing
    result = read_at_ix(structure_val, ix)
    #lhh
    #from pprint import pprint
    #pprint(structure_val)

    return result

#########################################################################
def list_exp(node):

    (LIST, inlist) = node
    assert_match(LIST, 'list')

    outlist =[]

    for e in inlist:
        outlist.append(walk(e))

    return ('list', outlist)

#########################################################################
def tuple_exp(node):

    (TUPLE, intuple) = node
    assert_match(TUPLE, 'tuple')

    outtuple = []

    for e in intuple:
        outtuple.append(walk(e))

    return ('tuple', outtuple)

#########################################################################
def escape_exp(node):

    (ESCAPE, s) = node
    assert_match(ESCAPE, 'escape')

    global __retval__
    __retval__ = ('none', None)

    exec(s)

    return __retval__

#########################################################################
def is_exp(node):

    (IS, term, pattern) = node
    assert_match(IS, 'is')

    term_val = walk(term)

    try:
        unifiers = unify(term_val, pattern)
    except PatternMatchFailed:
        return ('boolean', False)
    else:
        declare_unifiers(unifiers)
        return ('boolean', True)

#########################################################################
def in_exp(node):

    (IN, exp, exp_list) = node
    assert_match(IN, 'in')

    exp_val = walk(exp)
    (EXP_LIST_TYPE, exp_list_val, *_) = walk(exp_list)

    if EXP_LIST_TYPE != 'list':
        raise ValueError("right argument to 'in' operator has to be a list")

    # we simply map our in operator to the Python in operator
    if exp_val in exp_list_val:
        return ('boolean', True)
    else:
        return ('boolean', False)

#########################################################################
def if_exp(node):

    (IF_EXP, cond_exp, then_exp, else_exp) = node
    assert_match(IF_EXP, 'if-exp')

    # if expressions without an else clause are only allowed in
    # conditional patterns.
    if else_exp[0] == 'null':
        raise ValueError("if expressions need an 'else' clause")

    (BOOLEAN, cond_val) = map2boolean(walk(cond_exp))

    if cond_val:
        return walk(then_exp)
    else:
        return walk(else_exp)

#########################################################################
# NOTE: 'to-list' is not a semantic value and should never appear in
#       any tests.  It is a constructor and should be expanded by the
#       walk function before semantic processing.
def to_list_exp(node):

    (TOLIST,
     (START, start),
     (STOP, stop),
     (STEP, step)) = node

    assert_match(TOLIST, 'to-list')
    assert_match(START, 'start')
    assert_match(STOP, 'stop')
    assert_match(STEP, 'step')

    (START_TYPE, start_val, *_) = walk(start)
    (STOP_TYPE, stop_val, *_) = walk(stop)
    (STEP_TYPE, step_val, *_) = walk(step)

    if START_TYPE != 'integer' or STOP_TYPE != 'integer' or STEP_TYPE != 'integer':
        raise ValueError("only integer values allowed in start, stop, or step")

    out_list_val = []

    if int(step_val) > 0: # generate the list
        ix = int(start_val)
        while ix <= int(stop_val):
            out_list_val.append(('integer', ix))
            ix += int(step_val)

    elif int(step_val) == 0: # error
        raise ValueError("step size of 0 not supported")

    elif int(step_val) < 0: # generate the list
        ix = int(start_val)
        while ix >= int(stop_val):
            out_list_val.append(('integer', ix))
            ix += int(step_val)

    else:
        raise ValueError("{} not a valid step value".format(step_val))

    return ('list', out_list_val)

#########################################################################
# NOTE: this is the value view of the head tail constructor, for the
#       pattern view of this constructor see unify.
def head_tail_exp(node):

    (HEAD_TAIL, head, tail) = node
    assert_match(HEAD_TAIL, 'head-tail')

    head_val = walk(head)
    (TAIL_TYPE, tail_val) = walk(tail)

    if TAIL_TYPE != 'list':
        raise ValueError(
            "unsupported tail type '{}' in head-tail operator".
            format(TAIL_TYPE))

    return ('list', [head_val] + tail_val)

#########################################################################
# turn a function expression into a closure.
def function_exp(node):

    (FUNCTION_EXP, body_list) = node
    assert_match(FUNCTION_EXP,'function-exp')

    return ('function-val',
            body_list,
            state.symbol_table.get_closure())

#########################################################################
# Named patterns - when walking a named pattern we are interpreting a
# a pattern as a constructor - ignore the name
def named_pattern_exp(node):

    (NAMED_PATTERN, name, pattern) = node
    assert_match(NAMED_PATTERN,'named-pattern')

    return walk(pattern)

#########################################################################
def process_lineinfo(node):

    (LINEINFO, lineinfo_val) = node
    assert_match(LINEINFO, 'lineinfo')

    if debugging:
        debugger.set_lineinfo(lineinfo_val)

    #lhh
    #print("lineinfo: {}".format(lineinfo_val))

    state.lineinfo = lineinfo_val

#########################################################################
def deref_exp(node):

    (DEREF, id_exp) = node
    assert_match(DEREF, 'deref')

    # deref operators are only meaningful during pattern matching
    # ignore during a value walk.
    # NOTE: the second walk is necessary to interpret what we retrieved
    # through the indirection
    return walk(walk(id_exp))

#########################################################################
def constraint_exp(node):

    # Constraint-only pattern matches should not exist where only an
    # expression is expected. If we get here, we have come across this
    # situation.
    # A constraint-only pattern match AST cannot be walked and therefor
    # we raise an error.
    raise ValueError(
        "constraint pattern: '{}' cannot be used as a constructor."
        .format(term2string(node)))

#########################################################################
def set_ret_val(node):
    (SET_RET_VAL, exp) = node
    assert_match(SET_RET_VAL,'set-ret-val')

    global function_return_value
    val = walk(exp)
    function_return_value.pop()
    function_return_value.append(val)

    # Debugger keeps track of the most recent
    # return value
    if debugging and val:
        debugger.retval = term2string(val)

    return

#########################################################################
def walk_stmt_list(stmts, step_state=None):
    # step_stae is a flag to tell us if we actually
    # want to step through to the next line of the stmt
    # list while debugging
    
    # We only want to do this if we're debugging
    # and we've stepped through to this point or
    # have continued to this point. Basically, this
    # flag tells us if we've stepped or continued 
    # into this function call, effectively telling
    # the debugger to treat it as the top level

    # This also allows us to keep loop execution
    # control working with the debugger. Meaning
    # we can do a "next" on the last line in a 
    # loop and go back to the top as opposed to the
    # next line after the loop
    if step_state:
        stepping = step_state
    else:
        stepping = debugger_has_stepped()

    for s in stmts[1]:

        if debugging:
            # Set the debugger's info reflect new "top level"
            if stepping: debugger.set_top_level(True)

            # If we've hit a return and we're in continue-until-return
            # mode, notify the debugger

            # TODO: (OWM) get this working for implicit returns as well.
            # This currently doesn't work because you can have an implicit
            # return before a return statement. In this case, the return
            # statement consumes the return behavior because it is the
            # final statement in the stmt_list

            # OWM -- This is still true but I'm not sure the fix for it
            # nor if it should even be fixed

            # The statement is a return the debugger is executing until return
            # then notify the debugger
            if debugging and \
                debugger.exc['RETURN'] and s[0] == 'return':
                
                notify_debugger(at_return=True)
            
        walk(s)

#########################################################################
# walk
#########################################################################
def walk(node):
    # node format: (TYPE, [child1[, child2[, ...]]])
    type = node[0]

    if type == 'clear-ret-val':
        # implemented here instead of dictionary for efficiency reasons
        global function_return_value    

        # If we have no function return value, then append None    
        if function_return_value != []:
            function_return_value.pop()
        
        function_return_value.append(None)
        return
    elif type in dispatch_dict:
        node_function = dispatch_dict[type]
        return node_function(node)
    else:
        raise ValueError("feature '{}' not yet implemented".format(type))

#########################################################################
# walk_program
#########################################################################
def walk_program(node):
    """
    This function exists mostly as a performance
    boost to unify. Running the program through
    this function overrides the debugger wrapping
    on unify.
    """
    global unify
    unify = __unify

    walk(node)

#########################################################################
# debug_walk
#########################################################################
def debug_walk(node, dbg):
    """
    This function allows us to keep the top-level distinction
    at the program level.
    """
    global debugging, debugger, unify
    debugging, debugger = (True, dbg)

    unify = debug_unify

    (LIST, inlist) = node

    for e in inlist:
        # We want to differentiate between top level and nested
        # statements. So, we run the list of statements outright
        # so we can control when we know we are at the top level
        debugger.set_top_level(True)
        walk(e)

# a dictionary to associate tree nodes with node functions
dispatch_dict = {
    # statements - statements do not produce return values
    'lineinfo'      : process_lineinfo,
    'set-ret-val'   : set_ret_val,
    'noop'          : lambda node : None,
    'assert'        : assert_stmt,
    'unify'         : unify_stmt,
    'while'         : while_stmt,
    'loop'          : loop_stmt,
    'repeat'        : repeat_stmt,
    'for'           : for_stmt,
    'global'        : global_stmt,
    'return'        : return_stmt,
    'break'         : break_stmt,
    'if'            : if_stmt,
    'throw'         : throw_stmt,
    'try'           : try_stmt,
    'struct-def'    : struct_def_stmt,
    'top-level-exp' : top_level_exp_stmt,
    # expressions - expressions do produce return values
    'list'          : list_exp,
    'import_stmt'   : import_stmt,
    'tuple'         : tuple_exp,
    'to-list'       : to_list_exp,
    'head-tail'     : head_tail_exp,
    'raw-to-list'   : lambda node : walk(('to-list', node[1], node[2], node[3])),
    'raw-head-tail' : lambda node : walk(('head-tail', node[1], node[2])),
    'none'          : lambda node : node,
    'nil'           : lambda node : node,
    'function-exp'  : function_exp,
    'string'        : lambda node : node,
    'integer'       : lambda node : node,
    'real'          : lambda node : node,
    'boolean'       : lambda node : node,
    'object'        : lambda node : node,
    'eval'          : eval_exp,
    # pattern code should be treated like a constant if not ignore_pattern
    'pattern'         : lambda node : walk(node[1]) if state.ignore_pattern else node,
    # constraint patterns
    'constraint'    : constraint_exp,
    'typematch'     : constraint_exp,
    # type tag used in conjunction with escaped code in order to store
    # foreign objects in Asteroid data structures
    'foreign'       : lambda node : node,
    'id'            : lambda node : state.symbol_table.lookup_sym(node[1]),
    'apply'         : apply_exp,
    'index'         : index_exp,
    'escape'        : escape_exp,
    'is'            : is_exp,
    'in'            : in_exp,
    'if-exp'        : if_exp,
    'named-pattern' : named_pattern_exp,
    'member-function-val' : lambda node : node,
    'deref'         : deref_exp,
}

##############################################################################################
# *** The Redundant Pattern Detector ***
#
# Evaluates the presence of redundant, or 'useless', pattern clauses in an Asteroid function:
#
# A redundant, or 'useless', pattern is defined as a pattern which can never be matched
# due to a preceeding pattern consuming all intended pattern matches.
#
# Consider the following Asteroid function:
#
# function testFunction
#   with (x,y) do
#       return 1.
#   orwith (x,1) do
#      return 2.
#   end function.
#
# In the above function, the pattern (x,1) can never be reached as the preceeding pattern (x,y)
# will consume all intended matches. Therefore, it is redundant.
#
# Function check_redundancy takes in a functions body list during parsing.
# This body list contains a functions patterns along with the associated bodies for each
# pattern. This function then evaluates if patterns exist within the passed in function that
# are redundant. If so, a warning is printed to the console identifing the offending
# pattern(s)
#
################################################################################################
def check_redundancy( body_list, f_name ):

    #Node type assertions
    #or "Make sure we are walking down the right part of the tree"
    (BODY_LIST, function_bodies ) = body_list
    assert_match(BODY_LIST,'body-list')
    (LIST, bodies) = function_bodies
    assert_match(LIST,'list')

    #compare every pattern with the patterns that follow it
    for i in range(0, len(bodies), 2):
        # Process lineinfo
        lineinfo = bodies[ i ]
        process_lineinfo(lineinfo)

        #get the pattern with the higher level of precedence
        (BODY_H,(PTRN,ptrn_h),stmts_h) = bodies[i + 1]
        assert_match(BODY_H,'body')
        assert_match(PTRN,'pattern')

        for j in range(i + 2, len(bodies), 2):
            lineinfo = bodies[ j ]
            process_lineinfo(lineinfo)

            #get the pattern with the lower level of precedence
            (BODY_L,(PTRN,ptrn_l),stmts_l) = bodies[j + 1]
            assert_match(BODY_L,'body')
            assert_match(PTRN,'pattern')

            #Here we get line numbers in case we throw an error
            # we have to do a little 'tree walking' to get to the
            # line #, hence all the unpacking.
            (STMT_LIST,(LIST,LINE_LIST)) = stmts_l
            first_line_l = LINE_LIST[0]
            (LINE_INFO,location_l) = first_line_l

            (STMT_LIST,(LIST,LINE_LIST)) = stmts_h
            first_line_h = LINE_LIST[0]
            (LINE_INFO,location_h) = first_line_h

            # Compare the patterns to determine if the pattern with the
            # higher level of precedence will render the pattern with
            # the lower level of precedence useless/redundant by calling
            # on the unify function to evaluate the subsumption relationship
            # between the two patterns.
            try:                                #CHECK FOR CONFLICTION
                unify( ptrn_l, ptrn_h , False )
            except PatternMatchFailed:          #NO CONFLICTION
                pass
            else:                               #CONFLICTION
                raise RedundantPatternFound( ptrn_h , ptrn_l , f_name, location_h, location_l )

#######################################################################################
# *** Asteroid Debugger (ADB) helper functions ***
# 
# message_explicit:
#   The main messaging function for explicit mode. Operates in a similar way the
#   string.format function does. Read the function's internal docstring for more
#   information.
# 
# notify_debugger:
#   Notifies the debugger of the current program position. If the debugger is
#   accepting notifications at the time of call, the debugger will pause and
#   run the interactive prompt.
# 
# notify_explicit:
#   Similar to notify_debugger but only runs in explicit mode.
# 
# explicit_enabled:
#   A simple helper function. Equivalent to `debugging and debugger.explicit_enabled`
# 
# decrease_tab_level:
#   A simple helper function to manage debugger tab level decrementing.
#
# debugger_has_stepped::
#   Helper function that returns if the debugger has stepped into a function
#   body or other compound statement
#
# gen_t2s:
#   A generator function used in place of term2string in message_explicit calls.
#   This allows the computation to be deferred or completely ignored. Much like
#   a function returning a lambda function
#
# debug_unify:
#   A wrapper for the normal unify function that automatically decreases
#   the tab level of failed pattern matches.
#
#######################################################################################
def message_explicit(fmt_message, terms=None, level="primary", 
    increase=False, decrease=False, notify=False):

    """
    Message explicit expects 1-3 main arguments

    1. A string, which can be a format string or a normal string.
    2. (Optional) A list of terms to express for a given format string
    3. The message level (Defaults to primary)

    Format terms can be either plain strings or generators. If a gener-
    ator is supplied, this function will express it

    fmt_message is a list of terms to be applied in the same
    way as .format

    --------------------------------------------------------------
    This function also supports three other options that shorten
    common development patterns.

    increase: Increases the tab level after the message is written
    decrease: Decreases the tab level before the message is written
    notify: Makes a notify_explicit call to stop the debugger within
            a pattern
    """
    from types import GeneratorType

    if explicit_enabled():
        if decrease: decrease_tab_level()

        if not terms:
            debugger.message_explicit(fmt_message, level)
        else:
            expressed_terms = []
            for t in terms:
                if isinstance(t, GeneratorType):
                    expressed_terms.append(next(t))
                else:
                    expressed_terms.append(t)
            
            expressed_string = fmt_message.format(*expressed_terms)
            debugger.message_explicit(expressed_string, level)
        
        if increase:    debugger.tab_level += 1
        if notify:      notify_explicit()

#########################################################################
def notify_debugger(at_return=False):
    """
    If the debugger is accepting notifications at the time of call,
    the debugger will pause and run the interactive prompt.
    """
    if debugging:
        # We need to save the old lineinfo in case we go into a REPL
        old_lineinfo = state.lineinfo

        debugger.notify(at_return)

        # Reset our lineinfo
        state.lineinfo = old_lineinfo
        debugger.set_lineinfo(state.lineinfo)

#########################################################################
def notify_explicit():
    """
    If the debugger is accepting notifications at the time of call and
    the debugger is in explicit mode, the debugger will pause and run
    the interactive prompt.
    """
    if debugging and explicit_enabled:
        old_lineinfo = state.lineinfo
        
        debugger.notify_explicit()

        # Reset our lineinfo
        state.lineinfo = old_lineinfo
        debugger.set_lineinfo(state.lineinfo)

#########################################################################
def explicit_enabled():
    """
    Helper function that returns the state
    of explicit mode if debugging is enableds
    """
    return debugging and debugger.explicit_enabled

#########################################################################
def gen_t2s(node):
    """
    Generator function for term2string. This cuts down on
    runtime as it defers term2string computations to when
    they're actually needed.
    """
    yield term2string(node)

#########################################################################
def debugger_has_stepped():
    """
    Returns the step/continue state of the debugger
    """
    return debugging and \
            (debugger.exc['STEP'] or debugger.exc['CONTINUE'])

#########################################################################
def decrease_tab_level():
    """
    Decreases/handles the debugger's tab level
    """
    if debugging:
        debugger.tab_level = debugger.tab_level - 1

        if debugger.tab_level < 0:
            debugger.tab_level = 0

#########################################################################
def debug_unify(term, pattern, unifying = True):
    '''
    Proxy function for unify to make the debugger's formatting
    work properly

    OWM - This proxy function allows us to wrap unification
    in an except block and catch when PatternMatchFailed exceptions
    occur. In this situation, we need to decrease the tab level
    and then re-raise the exception. Being able to "unravel"
    tab levels is essential to making things look good.

    In normal, non-debugging execution, this function is overridden
    and not used.
    '''

    # Try to call the actual unify function
    try:
        return __unify(term, pattern, unifying)

    # If there's a pattern match failed, decrease
    # the tab level and reraise the exception
    except PatternMatchFailed as r:
        decrease_tab_level()
        raise r

    # If there's a normal exception, reraise it
    except Exception as e:
        raise e


#!/usr/bin/env python3
###########################################################################################
# Asteroid toplevel driver
#
# (c) University of Rhode Island
###########################################################################################

import cProfile
import sys
import os
from asteroid.interp import interp
from asteroid.repl import repl
from asteroid.version import VERSION
from asteroid.adb import adb

def display_help():
    print("** Asteroid Version {} **".format(VERSION))
    print("(c) University of Rhode Island")
    print("usage: asteroid [-<switch>] <input file>")
    print("")
    print("command line flags:")
    print(" -s             enable symbol table dump")
    print(" -t             AST dump")
    print(" -v, --version  version")
    print(" -w             disable tree walk")
    print(" -z             generate pstats")
    print(" -p             disable prologue")
    print(" -h, --help     display help")
    print(" -r             disable redundant pattern detector")
    print(" -e             show Python exceptions")
    print(" -F             functional mode")
    print(" -g, --adb      run program through debugger")

def main():
    # defaults for the flags - when the flag is set on the command line
    # it simply toggles the default value in this table.
    flags = {
        '-s' : False,  # symbol table dump flag
        '-t' : False,  # AST dump flag
        '--version' : False,  # version flag
        '-v' : False,  # version flag
        '-w' : True,   # tree walk flag
        '-z' : False,  # generate pstats flag
        '-p' : True,   # prologue flag
        '--help' : False,  # display help flag
        '-h' : False,  # display help flag
        '-r' : True,   # redundant pattern dectector
        '-e' : False,  # show full exceptions
        '-F' : False,  # functional mode
        '--adb': False, # debugger flag
        '-g': False    # Short debugger flag
    }

    flag_names = list(flags.keys())

    if len(sys.argv) == 1:
        repl(redundancy=flags['-r'],
             prologue=flags['-p'],
             functional_mode=flags['-F'])
        sys.exit(0)

    for fl in sys.argv:
        if fl[0] != '-':
            continue
        elif fl not in flag_names:
            print("unknown flag {}".format(fl))
            sys.exit(0)
        flags[fl] = not flags[fl]

    if flags['--help'] or flags['-h']:
        display_help()
        sys.exit(0)

    if flags['--version'] or flags['-v']:
        print("** Asteroid Version {} **".format(VERSION))
        sys.exit(0)

    input_file = sys.argv[-1]

    if input_file[0] == '-':
        if flags['--adb']:
            print("Please provide a file to debug")
            sys.exit(1)
        repl(redundancy=flags['-r'],
             prologue=flags['-p'],
             functional_mode=flags['-F'])
        sys.exit(0)

    if not os.path.isfile(input_file):
        print("unknown file {}".format(input_file))
        sys.exit(0)

    if flags['--adb'] or flags['-g']:
        # Create a new debugger
        db = adb.ADB()

        # Set the debugger's internal interpretation options
        db.interp_options = {
            'redundancy': flags['-r'],
            'prologue': flags['-p'],
            'functional_mode': flags['-F'],
            'exceptions': flags['-e'],
        }

        db.run(input_file)
        sys.exit(0)

    f = open(input_file, 'r')
    input_stream = f.read()
    f.close()

    # execute interpreter
    interp_object = \
    '''interp(program=input_stream,
           program_name = input_file,
           tree_dump=flags['-t'],
           do_walk=flags['-w'],
           symtab_dump=flags['-s'],
           exceptions=flags['-e'],
           redundancy=flags['-r'],
           prologue=flags['-p'],
           functional_mode=flags['-F'])'''

    if flags['-z']:
        # generates pstats into the file 'pstats'
        # see https://docs.python.org/3/library/profile.html
        cProfile.run(interp_object, 'pstats')
    else:
        exec(interp_object)

# for manual testing purposes
if __name__ == "__main__":
    main()



import re

class Token:
    def __init__(self,type,value):
        self.type = type
        self.value = value

    def __str__(self):
        return 'Token({},{})'.format(self.type,self.value)

class DebuggerLexer:
    def __init__(self, input_string):
        self.token_specs = [
        #   type:          value:
            ('STRING',     r'"(\\"|\\.|[^"])*"'),
            
            ('NEXT',        r'\bnext\b|\bn\b'),
            ('STEP',        r'\bstep\b|\bs\b'),
            ('CONTINUE',    r'\bcontinue\b|\bcont\b|\bc\b'),
            ('RETURN',      r'\breturn\b|\bret\b|\br\b'),
            ('UNTIL',       r'\buntil\b|\bu\b'),

            ('EVAL',        r'\beval\b'),
            ('BANG',        r'!'),

            ('BREAK',       r'\bbreak\b|\bb\b'),
            ('DELETE',      r'\bdelete\b|\bdel\b|\bd\b'),

            ('MACRO',       r'\bmacro\b'),
            ('LIST',        r'\blist\b|\bl\b'),
            ('LONGLIST',    r'\blonglist\b|\bll\b'),

            ('RETVAL',      r'\b__retval__\b|\b_\b'),
            
            ('EXPLICIT',    r'\bexplicit\b|\be\b'),
            ('ON',          r'\bon\b'),
            ('OFF',         r'\boff\b'),

            ('HELP',        r'\bh\b|\bhelp\b'),
            ('WHERE',       r'\bwhere\b|\bw\b'),
            ('UP',          r'\<'),
            ('DOWN',        r'\>'),

            ('IF',          r'\bif\b'),

            ('NUM',         r'[+-]?([0-9]*[.])?[0-9]+'),
            ('EQUAL',       r'='),
            ('SEMI',        r';'),
            ('COMMA',       r','),
            ('LPAREN',      r'\('),
            ('RPAREN',      r'\)'),

            ('QUIT',        r'\bquit\b|\bq\b'),

            ('NAME',        r'[a-zA-Z_\$][a-zA-Z0-9_\$]*'),
            ('WHITESPACE',  r'[ \t\n]+'),
            ('UNKNOWN',     r'.')
        ]

        # used for sanity checking in lexer.
        self.token_types = set(type for (type,_) in self.token_specs)
        self.tokens = self.tokenize(input_string)
        # the following is always valid because we will always have
        # at least the EOF token on the tokens list.
        self.curr_token_ix = 0

    def pointer(self):
        return self.tokens[self.curr_token_ix]

    def next(self):
        if not self.end_of_file():
            self.curr_token_ix += 1
        return self.pointer()

    def match(self, token_type):
        if token_type == self.pointer().type:
            tk = self.pointer()
            self.next()
            return tk
        elif token_type not in self.token_types:
            raise ValueError("unknown token type '{}'".format(token_type))
        else:
            raise ValueError('unexpected token {} while parsing, expected {}'
                              .format(self.pointer().type, token_type))

    def end_of_file(self):
        if self.pointer().type == 'EOF':
            return True
        else:
            return False

    def tokenize(self, code):
        tokens = []
        re_list = ['(?P<{}>{})'.format(type,re) for (type,re) in self.token_specs]
        combined_re = '|'.join(re_list)
        match_object_list = list(re.finditer(combined_re, code))
        for mo in match_object_list:
            type = mo.lastgroup
            value = mo.group()

            if type == 'WHITESPACE':
                continue #ignore
            elif type == 'STRING':
                tokens.append(Token('STRING', value.encode('utf-8')
                                                   .decode('unicode-escape')[1:-1]))
            elif type == 'UNKNOWN':
                raise ValueError("unexpected character '{}'".format(value))
            else: 
                tokens.append(Token(type, value))
        tokens.append(Token('EOF', '\eof'))
        return tokens

class DebuggerParser:
    def __init__(self):
        self.dlx = None

    def parse(self, input_string):
        self.dlx = DebuggerLexer(input_string)
        return self.line()

    def line(self):
        if self.dlx.pointer().type == 'MACRO':
            return self.macro()

        else:
            cmds = [self.command()]
            while self.dlx.pointer().type in ['SEMI']:
                self.dlx.match('SEMI')

                if self.dlx.pointer().type == 'EOF':
                    break

                cmds += [self.command()]

            return ('LINE', cmds)
    
    def macro(self):
        self.dlx.match('MACRO')

        if self.dlx.pointer().type == 'EOF':
            return ('LINE', [('MACRO',)])

        else:
            name = self.dlx.match('NAME')
            self.dlx.match('EQUAL')

            l = self.line()

            return ('LINE', [('MACRO', name.value, l[1])] )

    def eval_cmd(self):
        self.dlx.match("EVAL")
        self.dlx.match('LPAREN')
        code = self.dlx.match('STRING')
        self.dlx.match('RPAREN')
        
        return ('EVAL', code.value)
    
    def break_cmd(self):
        self.dlx.match('BREAK')
        nums = []
        conds = []
        first = True

        while self.dlx.pointer().type == 'NUM':
            if first:
                first = False
            
            nums.append(self.dlx.match('NUM').value)

            if self.dlx.pointer().type == 'IF':
                self.dlx.match('IF')
                self.dlx.match('EVAL')
                self.dlx.match('LPAREN')
                code = self.dlx.match('STRING')
                self.dlx.match('RPAREN')
                conds.append(code.value)
            else:
                conds.append(None)

            if not first:
                if self.dlx.pointer().type == 'COMMA':
                    self.dlx.match('COMMA')
                    continue
                elif self.dlx.pointer().type == 'EOF':
                    self.dlx.match('EOF')
                    break

        return ('BREAK', list(map(int, nums)), conds)

    def delete_cmd(self):
        self.dlx.match('DELETE')
        nums = [self.dlx.match('NUM').value]
        
        while self.dlx.pointer().type == 'NUM':
            nums.append(self.dlx.match('NUM').value)
        
        return ('DELETE', list(map(int, nums)))


    def help_cmd(self):
        self.dlx.match('HELP')
        
        n = None
        if self.dlx.pointer().type != 'EOF':
            n = self.dlx.pointer().value
            self.dlx.next()
        
        return ('HELP', n)

    def explicit_cmd(self):
        self.dlx.match('EXPLICIT')

        set_exp = None
        if self.dlx.pointer().type in ['ON', 'OFF']:
            set_exp = (self.dlx.pointer().type == 'ON')
            self.dlx.next()
        
        return ('EXPLICIT', set_exp)

    def until_cmd(self):
        self.dlx.match('UNTIL')
        
        n = None
        if self.dlx.pointer().type == 'NUM':
            n = self.dlx.pointer().value
            self.dlx.next()
        
        return ('UNTIL', n)

    def command(self):
        t = self.dlx.pointer().type
        
        if t ==  'EVAL':     return self.eval_cmd()
        elif t ==  'BREAK':    return self.break_cmd()
        elif t ==  'DELETE':   return self.delete_cmd()
        elif t ==  'HELP':     return self.help_cmd()
        elif t ==  'EXPLICIT': return self.explicit_cmd()
        elif t ==  'UNTIL':    return self.until_cmd()
        
        elif t in ['BANG', 'LONGLIST', 'LIST', 'QUIT', 'RETURN', 'RETVAL',
             'STEP', 'CONTINUE', 'NEXT', 'UP', 'DOWN', 'WHERE']:
            t = self.dlx.pointer().type
            self.dlx.match(t)
            return (t,)
        
        elif t ==  'NAME':
            n = self.dlx.match('NAME').value
            return ('NAME', n)
        
        elif t in  ['EOF', 'SEMI']:
            return ('NOOP',)
        
        else:
            raise ValueError("Unknown command: {}".format(
                str(self.dlx.pointer().value)
            ))


#########################################################################
# symbol table for Asteroid
#
# it is a scoped symbol table with a dictionary at each scope level
# each symbols is entered with a list of values in order to enable 'attach'
#
# (c) Lutz Hamel, University of Rhode Island
#########################################################################

from pprint import pprint

#########################################################################

CURR_SCOPE = 0

class SymTab:

    def __init__(self):
        # global scope dictionary must always be present
        self.scoped_symtab = [{}]
        self.globals = [[]]
        self.global_scope = self.scoped_symtab[0] # reference to global dictionary

        # Stack of configs not currently in use
        self.saved_configs = []

    def dump(self):
        print("*** Symbol Table Dump:")
        pprint(self.scoped_symtab)
        print("*** Globals Table:")
        pprint(self.globals)
        print("*** Global Scope:")
        pprint(self.global_scope)

    def get_closure(self):
        # we make a shallow copy of the symbol table, globals...
        return (list(self.scoped_symtab), list(self.globals), self.global_scope)

    def get_config(self):
        # return the relevant symtab objects
        return (self.scoped_symtab, self.globals, self.global_scope)

    def set_config(self, c):
        self.scoped_symtab, self.globals, self.global_scope = c

    def push_scope(self, scope):
        # push a new dictionary and globals lookup onto the stacks - stacks grow to the left
        self.scoped_symtab.insert(CURR_SCOPE, scope)
        self.globals.insert(CURR_SCOPE,[])

    def pop_scope(self):
        # pop the left most dictionary off the stack
        if len(self.scoped_symtab) == 1:
            raise ValueError("cannot pop the global scope")
        else:
            scope = self.scoped_symtab.pop(CURR_SCOPE)
            self.globals.pop(CURR_SCOPE)
            return scope

    def enter_sym(self, sym, value):
        # enter the symbol in the appropriate scope
        if sym in self.globals[CURR_SCOPE]:
            scope_dict = self.global_scope
        else:
            scope_dict = self.scoped_symtab[CURR_SCOPE]

        scope_dict[sym] = value

    def enter_global(self, sym):
        # enter the symbol in the global table at the appropriate scope
        self.globals[CURR_SCOPE].append(sym)

    def lookup_sym(self, sym, strict=True):
        dict = self.find_sym_dict(sym)
        if not dict:
            if strict:
                raise ValueError("'{}' is not defined".format(sym))
            else:
                return None
        else:
            return dict[sym]

    def update_sym(self, sym, value):
        # find the first occurence of sym
        # and update the associated value

        dict = self.find_sym_dict(sym)
        if not dict:
            raise ValueError("'{}' is not defined".format(sym))
        else:
            dict[sym] = value
            return

    def is_symbol_local(self, sym):
        if sym in self.scoped_symtab[CURR_SCOPE]:
            return True
        else:
            return False

    def is_global(self, sym):
        if sym in self.globals[CURR_SCOPE]:
            return True
        else:
            return False

    def find_sym_dict(self, sym):
        n_scopes = len(self.scoped_symtab)
        for scope in range(n_scopes):
            if sym in self.scoped_symtab[scope]:
                return self.scoped_symtab[scope]
        # not found
        return None

    def dbg_find_sym(self, sym):
        n_scopes = len(self.scoped_symtab)
        for scope in range(n_scopes):
            if sym in self.scoped_symtab[scope]:
                return ("---- found symbol {} in scope {} with value {}"
                      .format(sym, scope, self.scoped_symtab[scope].get(sym)))
        # not found
        return ("{} was not found".format(sym))



#########################################################################



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="asteroid-lang",
    version="1.1.3",
    author="University of Rhode Island",
    author_email="lutzhamel@uri.edu",
    description="A modern, multi-paradigm programming language.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://asteroid-lang.org",
    project_urls={
        "Documentation": "https://asteroid-lang.readthedocs.io",
        "Bug Tracker":   "https://github.com/asteroid-lang/asteroid/issues",
    },
    packages=find_packages(),
    package_data={"asteroid": ["modules/*"]},
    install_requires=[
        "pandas",
        "numpy",
        "matplotlib"
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.6",
    entry_points={
        "console_scripts": [
            "asteroid = asteroid:main",
        ],
    },
)
